# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from warnings import warn
from typing import Sequence, Iterator
from traits.api import Undefined

from hyperspy._signals.lazy import LazySignal
from hyperspy.signal import BaseSignal
from hyperspy.axes import BaseDataAxis
from orix.crystal_map import CrystalMap, Phase, PhaseList
from orix.quaternion import Rotation, Orientation

from transforms3d.euler import mat2euler
from diffsims.crystallography._diffracting_vector import DiffractingVector

from orix.vector import Vector3d
from orix.projections import StereographicProjection
from orix.plot.inverse_pole_figure_plot import _get_ipf_axes_labels
from orix.vector.fundamental_sector import _closed_edges_in_hemisphere
from orix.plot import IPFColorKeyTSL, DirectionColorKeyTSL

from matplotlib.collections import QuadMesh
from matplotlib.colors import Normalize
from scipy.spatial import Delaunay

import numpy as np
import hyperspy.api as hs

from pyxem.utils.indexation_utils import get_nth_best_solution, dict2phase, phase2dict
from pyxem.signals.diffraction_vectors2d import DiffractionVectors2D
from pyxem.utils._signals import _transfer_navigation_axes
from pyxem.utils.signal import compute_markers
from pyxem.utils._deprecated import deprecated


def crystal_from_vector_matching(z_matches):
    """Takes vector matching results for a single navigation position and
    returns the best matching phase and orientation with correlation and
    reliability to define a crystallographic map.

    Parameters
    ----------
    z_matches : numpy.ndarray
        Template matching results in an array of shape (m,5) sorted by
        total_error (ascending) within each phase, with entries
        [phase, R, match_rate, ehkls, total_error]

    Returns
    -------
    results_array : numpy.ndarray
        Crystallographic mapping results in an array of shape (3) with entries
        [phase, np.array((z, x, z)), dict(metrics)]

    """
    if z_matches.shape == (1,):  # pragma: no cover
        z_matches = z_matches[0]

    # Create empty array for results.
    results_array = np.empty(3, dtype="object")

    # get best matching phase
    best_match = get_nth_best_solution(
        z_matches, "vector", key="total_error", descending=False
    )
    results_array[0] = best_match.phase_index

    # get best matching orientation Euler angles
    results_array[1] = np.rad2deg(mat2euler(best_match.rotation_matrix, "rzxz"))

    # get vector matching metrics
    metrics = dict()
    metrics["match_rate"] = best_match.match_rate
    metrics["ehkls"] = best_match.error_hkls
    metrics["total_error"] = best_match.total_error

    results_array[2] = metrics

    return results_array


def _get_best_match(z):
    """Returns the match with the highest score for a given navigation pixel

    Parameters
    ----------
    z : numpy.ndarray
        array with shape (5,n_matches), the 5 elements are phase, alpha, beta, gamma, score

    Returns
    -------
    z_best : numpy.ndarray
        array with shape (5,)

    """
    return z[np.argmax(z[:, -1]), :]


def _get_phase_reliability(z):
    """Returns the phase reliability (phase_alpha_best/phase_beta_best) for a given navigation pixel

    Parameters
    ----------
    z : numpy.ndarray
        array with shape (5,n_matches), the 5 elements are phase, alpha, beta, gamma, score

    Returns
    -------
    phase_reliabilty : float
        np.inf if only one phase is avaliable
    """
    best_match = _get_best_match(z)
    phase_best = best_match[0]
    phase_best_score = best_match[4]

    # mask for other phases
    lower_phases = z[z[:, 0] != phase_best]
    # needs a second phase, if none return np.inf
    if lower_phases.size > 0:
        phase_second = _get_best_match(lower_phases)
        phase_second_score = phase_second[4]
    else:
        return np.inf

    return phase_best_score / phase_second_score


def _get_second_best_phase(z):
    """Returns the the second best phase for a given navigation pixel

    Parameters
    ----------
    z : numpy.ndarray
        array with shape (5,n_matches), the 5 elements are phase, alpha, beta, gamma, score

    Returns
    -------
    phase_id : int
        associated with the second best phase
    """
    best_match = _get_best_match(z)
    phase_best = best_match[0]

    # mask for other phases
    lower_phases = z[z[:, 0] != phase_best]

    # needs a second phase, if none return -1
    if lower_phases.size > 0:
        phase_second = _get_best_match(lower_phases)
        return phase_second[4]
    else:
        return -1


def get_ipf_outline(
    phase: Phase,
    include_labels: bool = True,
    offset_x: float = 0.85,
    offset_y: float = 0.85,
    scale: float = 0.2,
):
    """Get the outline of the IPF for the orientation map as a marker in the
    upper right hand corner including labels if desired.

    Parameters
    ----------
    phase: Phase
        The phase to use, defining the symmerty reduced zone
    include_labels : bool
        If True, the labels for the axes will be included.
    offset : float
        The offset of the markers from the lower left of the plot (as a fraction of the axis).
    scale : float
        The scale (as a fraction of the axis) for the markers.

    Returns
    -------
    polygon_sector : hs.plot.markers.Polygons
        The outline of the IPF as a marker
    texts : hs.plot.markers.Texts
        The text labels for the IPF axes
    maxes : np.ndarray
        The maximum values for the axes
    mins : np.ndarray
        The minimum values for the axes
    """
    offset = np.array([offset_x, offset_y])
    # Creating Lines around QuadMesh
    sector = phase.point_group.fundamental_sector
    s = StereographicProjection()

    edges = _closed_edges_in_hemisphere(sector.edges, sector)
    ex, ey = s.vector2xy(edges)
    original_offset = np.vstack((ex, ey)).T
    mins, maxes = original_offset.min(axis=0), original_offset.max(axis=0)
    original_offset = (
        (original_offset - ((maxes + mins) / 2)) / (maxes.max() - mins.min()) * scale
    )
    original_offset = original_offset + offset
    polygon_sector = hs.plot.markers.Polygons(
        verts=original_offset[np.newaxis],
        transform="axes",
        alpha=1,
        facecolor="none",
    )
    if include_labels:
        labels = _get_ipf_axes_labels(sector.vertices, symmetry=phase.point_group)
        tx, ty = s.vector2xy(sector.vertices)
        texts_offset = np.vstack((tx, ty)).T
        texts_offset = (
            (texts_offset - ((maxes + mins) / 2)) / (maxes.max() - mins.min()) * scale
        )
        texts_offset = texts_offset + offset
        texts = hs.plot.markers.Texts(
            texts=labels,
            offsets=texts_offset,
            sizes=(1,),
            offset_transform="axes",
            facecolor="k",
        )
        return polygon_sector, texts, maxes, mins


def get_ipf_annotation_markers(phase: Phase, offset: float = 0.85, scale: float = 0.2):
    """Get the outline of the IPF for the orientation map as a marker in the
    upper right hand corner including labels if desired. As well as the color
    mesh for the IPF.

    Parameters
    ----------
    offset : float
        The offset of the markers from the lower left of the plot (as a fraction of the axis).
    scale : float
        The scale (as a fraction of the axis) for the markers.

    Returns
    -------
    polygon_sector : hs.plot.markers.Polygons
        The outline of the IPF as a marker
    texts : hs.plot.markers.Texts
        The text labels for the IPF axes
    mesh : hs.plot.markers.Markers
        The color mesh for the IPF (using :class:`matplotlib.collections.QuadMesh`)
    """

    polygon_sector, texts, _, _ = get_ipf_outline(
        phase, offset_x=offset, offset_y=offset, scale=scale
    )

    # Create Color Mesh
    color_key = DirectionColorKeyTSL(symmetry=phase.point_group)
    g, ext = color_key._create_rgba_grid(return_extent=True)

    max_x = np.max(ext[1])
    min_x = np.min(ext[1])

    max_y = np.max(ext[0])
    min_y = np.min(ext[0])

    # center extent:
    y = np.linspace(ext[0][0], ext[0][1], g.shape[1] + 1) - ((max_y + min_y) / 2)

    y = y / (max_y - min_y) * scale + offset

    x = np.linspace(ext[1][1], ext[1][0], g.shape[0] + 1) - ((max_x + min_x) / 2)

    x = x / (max_x - min_x) * scale + offset
    xx, yy = np.meshgrid(y, x)

    mesh = hs.plot.markers.Markers(
        collection=QuadMesh,
        coordinates=np.stack((xx, yy), axis=-1),
        array=g,
        transform="axes",
        offset_transform="display",
        offsets=[[0, 0]],
    )
    return polygon_sector, mesh, texts


def vectors_to_single_phase_ipf_markers(
    vectors: Vector3d,
    phase: Phase,
    normalized_correlation: np.ndarray,
    offset_x: float = 0.85,
    offset_y: float = 0.85,
    scale: float = 0.2,
    color: str = "green",
    cmap: str = None,
):
    """Convert the orientation map to a set of inverse pole figure
    markers which visualizes the best matching orientations in the
    reduced S2 space.

    Parameters
    ----------
    vectors : Vector3d
        The vectors to display. These will be projected into the fundamental sector
    phase : Phase
        The phase for the fundamental sector
    normalized_correrlation : np.ndarray
        Correlation scores, normalized to lie between 0 and 1.
        Used for alpha for single-color plotting, and for the color map
    offset : float
        The offset of the markers from the center of the plot
    scale : float
        The scale (as a fraction of the axis) for the markers.
    """
    offset = np.array([offset_x, offset_y])

    markers = []

    vectors = vectors.in_fundamental_sector(phase.point_group)
    polygon_sector, texts, maxes, mins = get_ipf_outline(
        phase, offset_x=offset_x, offset_y=offset_y, scale=scale
    )
    markers.append(polygon_sector)
    markers.append(texts)

    s = StereographicProjection()
    x, y = s.vector2xy(vectors)
    x = x.reshape(vectors.shape)
    y = y.reshape(vectors.shape)
    offsets = np.empty(shape=vectors.shape[:-1], dtype=object)
    alpha = np.ones(shape=vectors.shape[:-1], dtype=object)

    # For each set of n_best, we reverse the order (using [::-1])
    # to ensure the best spot is displayed on top
    for i in np.ndindex(offsets.shape):
        off = np.vstack((x[i], y[i])).T
        norm_points = (off - ((maxes + mins) / 2)) / (maxes.max() - mins.min()) * scale
        norm_points = norm_points + offset
        offsets[i] = norm_points[::-1]
        alpha[i] = normalized_correlation[i][::-1]

    square = hs.plot.markers.Squares(
        offsets=[offset],
        widths=(scale + scale / 2,),
        units="width",
        offset_transform="axes",
        facecolor="white",
        edgecolor="black",
    )
    # Square needs to be on the bottom
    markers.insert(0, square)

    best_points = hs.plot.markers.Points(
        offsets=offsets.T,
        sizes=(4,),
        offset_transform="axes",
        alpha=alpha.T,
        facecolor=color,
        cmap=cmap,
        array=None if cmap is None else alpha.T,
        norm=Normalize(0, 1),
    )
    markers.append(best_points)

    return (*markers,)


def vectors_to_coordinates(vectors):
    """
    Convert a set of diffraction vectors to coordinates. For use with the map
    function and making markers.
    """
    return np.vstack((vectors[:, 0], vectors[:, 1])).T


def vectors_to_intensity(vectors, scale=1):
    """
    Convert a set of diffraction vectors to coordinates. For use with the map
    function and making markers.
    """

    return (vectors[:, 3] / np.max(vectors[:, 3])) * scale


def vectors_to_text(vectors, fast=True):
    """
    Convert a set of diffraction vectors to text. For use with the map function
    and making text markers.
    """

    def add_bar(i: int) -> str:
        if i < 0:
            return f"$\\bar{{{abs(i)}}}$"
        else:
            return f"{i}"

    out = []
    for hkl in vectors[:, 4:]:
        h, k, l = np.round(hkl).astype(np.int16)
        if fast:
            out.append(f"{h} {k} {l}")
        else:
            out.append(f"({add_bar(h)} {add_bar(k)} {add_bar(l)})")
    return out


def rotation_from_orientation_map(result, rots):
    if rots.ndim == 1:
        rots = rots[np.newaxis, :]
    index, _, rotation, mirror = result.T
    index = index.astype(int)
    ori = rots[index]
    euler = (
        Orientation(ori).to_euler(
            degrees=True,
        )
        * mirror[..., np.newaxis]
    )
    euler[:, 0] = rotation
    ori = Orientation.from_euler(euler, degrees=True).data
    return ori


@deprecated(
    since="0.20",
    removal="1.0.0",
    alternative="pyxem.signals.indexation_results.vectors_from_orientation_map",
)
def extract_vectors_from_orientation_map(**kwargs):
    return vectors_from_orientation_map(**kwargs)


def vectors_from_orientation_map(
    result,
    vectors,
    phases,
    phase_index,
    hkl,
    intensities,
    n_best_index=0,
    return_object=True,
):
    """
    Note that any function which passes a large number of python objects back/forth to dask causes
    lots of overhead.  We are going to try to stick with just passing numpy arrays... Also,
    the Orix Rotation class has a really high overhead which is unfortunate. That is something to
    consider optimizing in the future.

    Parameters
    ----------
    result : np.ndarray
        The result array from the orientation map
    vectors : DiffractingVector
        The vectors from the simulation
    phases : List of orix.crystal_map.Phase
        The phase array from the simulation
    phase_index : np.ndarray
        The index of the phase for each vector
    intensities : np.ndarray
        The intensities of the vectors
    n_best_index : int
        The index of the best match to use
    return_object : bool
        If True, return a DiffractingVector object, otherwise return a numpy array with a stack of
        [x, y, z, intensity, h, k, l]
    """
    index, _, rotation, mirror = result[n_best_index, :].T
    index = index.astype(int)
    if vectors.ndim == 0:
        data = vectors
        phase = phases
        hkl = hkl
    else:
        data = vectors[index]
        phase = phases[phase_index[index]]
        intensities = intensities[index]
        hkl = hkl[index]
    phase = dict2phase(**phase)
    # Copy manually, as deepcopy adds a lot of overhead with the phase
    vectors = DiffractingVector(phase, xyz=data.copy())
    # Flip y, as discussed in https://github.com/pyxem/pyxem/issues/925
    vectors.y = -vectors.y

    rotation = Rotation.from_euler(
        (mirror * rotation, 0, 0), degrees=True, direction="crystal2lab"
    )
    coordinate_format = vectors.coordinate_format
    vectors = ~rotation * vectors.to_miller()
    vectors = DiffractingVector(
        vectors.phase,
        xyz=-vectors.data,  # Negating for proper alignment - is this flipping z direction?
        intensity=intensities,
    )
    vectors.coordinate_format = coordinate_format
    vectors.original_hkl = hkl

    # angle = np.deg2rad(mirror * rotation)
    # rot_coj = np.array([np.cos(0.5 * angle), 0, 0, -np.sin(0.5 * angle)])
    # data = qu_rotate_vec(rot_coj, vectors.data)
    # vectors = DiffractingVector(vectors.phase, xyz=data, intensity=intensities)
    # vectors.original_hkl = hkl

    # Mirror if necessary.
    # Mirroring, in this case, means casting (r, theta) to (r, -theta).
    # This is equivalent to (x, y) -> (x, -y)
    vectors.y = mirror * vectors.y

    if return_object:
        return vectors
    else:
        return np.hstack((vectors.data, vectors.intensity[:, np.newaxis], hkl))


def orientation2phase(orientations, sizes):
    o = orientations[:, 0]
    return np.searchsorted(sizes, o)


class OrientationMap(DiffractionVectors2D):
    """Signal class for orientation maps.  Note that this is a special case where
    for each navigation position, the signal contains the top n best matches in the form
    of a nx4 array with columns [index,correlation, in-plane rotation, mirror(factor)]

    The Simulation is saved in the metadata but can be accessed using the .simulation attribute.

    Parameters
    ----------
    *args
        See :class:`~hyperspy._signals.signal2d.Signal2D`.
    **kwargs
        See :class:`~hyperspy._signals.signal2d.Signal2D`
    """

    _signal_type = "orientation_map"

    @property
    def simulation(self):
        """Simulation object used to generate the orientation map.  This is stored in the metadata
        but can be accessed using the ``.simulation`` attribute. The simulation object is a
        :class:`diffsims.simulation.Simulation2D` object."""
        return self.metadata.get_item("simulation")

    @simulation.setter
    def simulation(self, value):
        self.metadata.set_item("simulation", value)

    def deepcopy(self, deepcopy_simulation=False):
        """Deepcopy the signal. Deepcopying the simulation is optional and is computationally
        expensive. If the simulation is not deepcopied, the simulation attribute of the
        copied signal is passed along"""

        if not deepcopy_simulation:
            simulation = self.simulation
            simulation._phase_slider = None
            simulation._rotation_slider = None
            self.simulation = None
        new = super().deepcopy()
        if not deepcopy_simulation:
            new.simulation = simulation
            self.simulation = simulation
        return new

    @property
    def num_rows(self):
        return self.axes_manager.signal_axes[1].size

    def to_rotation(self, flatten=False):
        """
        Convert the orientation map to a set of `orix.Quaternion.Rotation` objects.

        Parameters
        ----------
        flatten : bool
            If True, the rotations will be flattened to a 2D array. This is useful for passing
            to a `CrystalMap` object.
        Returns
        -------
        orix.Quaternion.Rotation
        """
        if self._lazy:
            raise ValueError(
                "Cannot create rotation from lazy signal. Please compute the signal first."
            )
        if isinstance(self.simulation.rotations, (list, np.ndarray)):
            quats = np.vstack([r.data for r in self.simulation.rotations])
            all_rotations = Rotation(data=quats)
        else:
            all_rotations = self.simulation.rotations
        rotations = self.map(
            rotation_from_orientation_map,
            rots=all_rotations.data,
            inplace=False,
            lazy_output=False,
            output_signal_size=(self.num_rows, 4),
            output_dtype=float,
            silence_warnings=True,
        )

        rots = Rotation(rotations)
        if flatten:
            shape1 = np.prod(rots.shape[:-1])
            rots = rots.reshape((shape1, rots.shape[-1]))
        return rots

    def to_phase_index(self):
        """
        Convert the orientation map to a set of phase ids

        Returns
        -------
        np.ndarray or None
            The phase ids for each pixel in the orientation map or None if the simulation
            does not have multiple phases.
        """
        if self.simulation.has_multiple_phases:
            sizes = np.cumsum([i.size for i in self.simulation.rotations])
            if self.num_rows == 1:
                output_size = ()
            else:
                output_size = (self.num_rows,)
            return self.map(
                orientation2phase,
                sizes=sizes,
                inplace=False,
                output_signal_size=output_size,
                silence_warnings=True,
            ).data
        else:
            return None

    def to_single_phase_orientations(self, **kwargs) -> Orientation:
        """Convert the orientation map to an `Orientation`-object,
        given a single-phase simulation.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to pass to the :meth:`hyperspy.api.signals.BaseSignal.map` function.
        """
        if self.simulation.has_multiple_phases:
            raise ValueError(
                "Multiple phases found in simulation (use to_crystal_map instead)"
            )

        # Use the quaternion data from rotations to support 2D rotations,
        # i.e. unique rotations for each navigation position
        rotations = hs.signals.Signal2D(self.simulation.rotations.data)

        return Orientation(
            self.map(
                rotation_from_orientation_map,
                rots=rotations,
                inplace=False,
                output_signal_size=(self.num_rows, 4),
                output_dtype=float,
                silence_warnings=True,
                **kwargs,
            ),
            symmetry=self.simulation.phases.point_group,
        )

    @deprecated(
        since="0.20",
        removal="1.0.0",
        alternative="pyxem.signals.OrientationMap.to_vectors",
    )
    def to_single_phase_vectors(
        self, n_best_index: int = 0, **kwargs
    ) -> hs.signals.Signal1D:
        return self.to_vectors(n_best_index=n_best_index, **kwargs)

    def to_vectors(
        self, n_best_index: int = 0, return_object: bool = True, **kwargs
    ) -> hs.signals.Signal1D:
        """Get the reciprocal lattice vectors for each navigation position.

        Parameters
        ----------
        n_best_index: int
            The index into the `n_best` matchese
        **kwargs
            Additional keyword arguments to pass to the :meth:`hyperspy.api.signals.BaseSignal.map` function.
        """

        if self.simulation.has_multiple_phases:
            # With different phases, we cannot have a signal anymore, as the first dimension is now the phase
            vectors_signal = np.array([sim for sim in self.simulation])
        else:
            # Use vector data as signal in case of different vectors per navigation position
            vectors_signal = hs.signals.Signal1D(self.simulation.coordinates)

        data, intensities, phases, phase_indices, hkl = self.get_simulation_arrays()

        phases_dicts = [phase2dict(p) for p in phases]
        v = self.map(
            vectors_from_orientation_map,
            vectors=data,
            phases=phases_dicts,
            hkl=hkl,
            phase_index=phase_indices,
            intensities=intensities,
            inplace=False,
            output_signal_size=(),
            output_dtype=object,
            show_progressbar=False,
            n_best_index=n_best_index,
            return_object=return_object,
            silence_warnings=True,
            **kwargs,
        )
        v.metadata.simulation = None
        return v

    def get_simulation_arrays(self):
        data = np.array([sim.data for sim in self.simulation], dtype=object)
        hkl = np.array([sim.hkl for sim in self.simulation], dtype=object)
        intensities = np.array([sim.intensity for sim in self.simulation], dtype=object)
        if self.simulation.has_multiple_phases:
            phases = self.simulation.phases
            csum = np.append(
                [
                    0,
                ],
                np.cumsum([r.size for r in self.simulation.rotations]),
            )
            phase_indices = np.zeros(csum[-1], dtype=int)
            for c in range(len(self.simulation.rotations)):
                phase_indices[csum[c] : csum[c + 1]] = c
        else:
            phases = [
                self.simulation.phases,
            ]
            phase_indices = np.zeros(self.simulation.rotations.size, dtype=int)
        return data, intensities, phases, phase_indices, hkl

    def to_crystal_map(self) -> CrystalMap:
        """Convert the orientation map to an :class:`orix.crystal_map.CrystalMap` object

        Returns
        -------
        orix.crystal_map.CrystalMap
            A crystal map object with the phase id as the highest matching phase.
        """
        if self.axes_manager.navigation_dimension != 2:
            raise ValueError(
                "Only 2D navigation supported. Please raise an issue if you are interested in "
                "support for 3+ navigation dimensions."
            )

        x, y = [ax.axis for ax in self.axes_manager.navigation_axes]
        xx, yy = np.meshgrid(x, y)
        xx = xx - np.min(xx)
        yy = yy - np.min(yy)
        scan_unit = self.axes_manager.navigation_axes[0].units
        rotations = self.to_rotation(flatten=True)
        phase_index = self.to_phase_index()

        if self.simulation.has_multiple_phases:
            phases = PhaseList(list(self.simulation.phases))
            if phase_index.ndim == 3:
                phase_index = phase_index[..., 0]
            phase_index = phase_index.flatten()
        else:
            phases = PhaseList(self.simulation.phases)
        if scan_unit is Undefined:
            scan_unit = "px"

        return CrystalMap(
            rotations=rotations,
            x=xx.flatten(),
            phase_id=phase_index,
            y=yy.flatten(),
            scan_unit=scan_unit,
            phase_list=phases,
        )

    def to_ipf_markers(
        self,
        offset_x: float = 0.85,
        offset_y: float = 0.85,
        scale: float = 0.2,
        color: str = "green",
        cmap: str = None,
    ):
        """Convert the orientation map to a set of inverse pole figure
        markers which visualizes the best matching orientations in the
        reduced S2 space.

        Parameters
        ----------
        offset : float
            The offset of the markers from the center of the plot
        scale : float
            The scale (as a fraction of the axis) for the markers.
        color : str
            The color of the markers. Overridden by cmap, if not None
        cmap : str
            Use a color map to show correlation scores.
            Takes priority over the `color` parameter.

        Notes
        -----
        This function can be slow with a large n_keep.
        Use `to_ipf_correlation_heatmap_markers` instead.
        """
        rots = self.to_rotation()
        vecs = rots * Vector3d.zvector()
        # Normalize the correlation scores
        cors = self.data[..., 1] / np.max(self.data[..., 1], axis=-1)[..., np.newaxis]
        if not self.simulation.has_multiple_phases:
            return vectors_to_single_phase_ipf_markers(
                vecs,
                self.simulation.phases,
                cors,
                offset_x=offset_x,
                offset_y=offset_y,
                scale=scale,
                color=color,
                cmap=cmap,
            )
        # Multiple phases, we make a few different IPFs
        markers = []
        phase_idxs = self.to_phase_index()
        for phase_idx, phase in enumerate(self.simulation.phases):
            # As a workaround, set alpha to 0 for the wrong phase
            phase_cors = cors * (phase_idxs == phase_idx)
            markers += list(
                vectors_to_single_phase_ipf_markers(
                    vecs,
                    phase,
                    phase_cors,
                    offset_x=offset_x,
                    offset_y=offset_y - 1.5 * scale * phase_idx,
                    scale=scale,
                    color=color,
                    cmap=cmap,
                )
            )
        return markers

    def to_ipf_correlation_heatmap_markers(
        self,
        offset_x: float = 0.85,
        offset_y: float = 0.85,
        scale: float = 0.2,
        cmap: str = "inferno",
    ):
        """Get the outline of the IPF for the orientation map as a marker in the
        upper right hand corner including labels if desired. As well as the color
        mesh for the IPF.

        Parameters
        ----------
        offset : float
            The offset of the markers from the lower left of the plot (as a fraction of the axis).
        scale : float
            The scale (as a fraction of the axis) for the markers.

        Returns
        -------
        polygon_sector : hs.plot.markers.Polygons
            The outline of the IPF as a marker
        texts : hs.plot.markers.Texts
            The text labels for the IPF axes
        mesh : hs.plot.markers.Markers
            The color mesh for the IPF (using :class:`matplotlib.collections.QuadMesh`)

        Notes
        -----
        It is recommended to have a large `n_keep` for this to look nice, e.g. 75% of the simulation bank.

        This will not look good if the rotations used in the simulation(s) consists of
        multiple different regions in the IPF.

        """
        phase_idx_signal = hs.signals.Signal1D(self.to_phase_index())
        phases = self.simulation.phases
        rotations = self.simulation.rotations
        if not self.simulation.has_multiple_phases:
            phases = [phases]
            rotations = [rotations]
            phase_idx_signal = 0

        s = StereographicProjection()
        rotation_sizes = np.cumsum([0] + [r.size for r in rotations])

        markers = []

        for phase_idx, phase in enumerate(phases):
            offset = np.array([offset_x, offset_y])
            polygon_sector, texts, maxs, mins = get_ipf_outline(
                phase, offset_x=offset_x, offset_y=offset_y, scale=scale
            )
            # We make a mesh of the correlation scores,
            # But we mask out the <n_best rotations.
            # Fetch all possible zone axes for the phase to speed up calculations
            rots = rotations[phase_idx]
            vecs = rots * Vector3d.zvector()

            x, y = s.vector2xy(vecs)
            xy = np.vstack((x, y)).T

            new_x = np.linspace(mins[0], maxs[0], 100 + 1)
            new_y = np.linspace(mins[1], maxs[1], 100 + 1)
            xx, yy = np.meshgrid(new_x, new_y)
            out_xy = np.stack((xx, yy), axis=-1)
            new_xy = np.vstack((xx[:-1, :-1].flatten(), yy[:-1, :-1].flatten())).T

            # Interpolate the values to a grid.
            # This is usually slow, e.g. using scipy.interpolate.griddata.
            # Since the data coordinates and interpolated coordinates
            # are always the same, and since we use linear interpolation,
            # we can pre-calculate the barycentric weights for Delaunay triangles
            # https://stackoverflow.com/a/20930910
            tri = Delaunay(xy, incremental=False)
            simplex = tri.find_simplex(new_xy)

            # points outside the convex hull gets assigned the index -1.
            # We later mask these out
            outside = simplex < 0

            vertices = np.take(tri.simplices, simplex, axis=0)
            temp = np.take(tri.transform, simplex, axis=0)
            delta = new_xy - temp[:, -1]
            bary = np.einsum("njk,nk->nj", temp[:, :-1, :], delta)
            weights = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))

            def interpolate(values):
                return np.einsum("nj,nj->n", np.take(values, vertices), weights)

            def interpolate_and_mask_correlations(result, phase_idxs):
                indices, correlation, _, _ = result[phase_idxs == phase_idx].T
                max_cor = result[:, 1].max()
                indices = indices.astype(int) - rotation_sizes[phase_idx] - 1
                all_correlations = np.full(x.shape, np.nan)
                all_correlations[indices] = correlation / max_cor
                out = interpolate(all_correlations)
                out[outside] = np.nan
                return out

            g = self.map(
                interpolate_and_mask_correlations,
                phase_idxs=phase_idx_signal,
                inplace=False,
                ragged=True,
                silence_warnings=True,
            ).data

            # Scale the coordinates
            out_xy = (out_xy - ((maxs + mins) / 2)) / (maxs.max() - mins.min()) * scale
            out_xy += offset

            mesh = hs.plot.markers.Markers(
                collection=QuadMesh,
                coordinates=out_xy,
                array=g.T,
                transform="axes",
                offset_transform="display",
                offsets=[[0, 0]],
                norm=Normalize(0, 1),
                cmap=cmap,
            )
            square = hs.plot.markers.Squares(
                offsets=[offset],
                widths=(scale + scale / 2,),
                units="width",
                offset_transform="axes",
                facecolor="white",
                edgecolor="black",
            )

            markers.append(square)
            markers.append(polygon_sector)
            markers.append(mesh)
            markers.append(texts)

            offset_y -= 1.5 * scale
        return (*markers,)

    def to_markers(
        self,
        n_best: int = 1,
        annotate: bool = False,
        marker_colors: str = ("red", "blue", "green", "orange", "purple"),
        text_color: str = "black",
        lazy_output: bool = None,
        annotation_shift: Sequence[float] = None,
        text_kwargs: dict = None,
        include_intensity: bool = False,
        intensity_scale: float = 1,
        fast: bool = True,
        **kwargs,
    ) -> Sequence[hs.plot.markers.Markers]:
        """Convert the orientation map to a set of markers for plotting.

        Parameters
        ----------
        n_best: int
            The amount of solutions to plot
        annotate : bool
            If True, the euler rotation and the correlation will be annotated on the plot using
            the `Texts` class from hyperspy.
        marker_color: str, optional
            The color of the point markers used for simulated reflections
        text_color: str, optional
            The color used for the text annotations for reflections. Does nothing if `annotate` is `False`.
        annotation_shift: List[float,float], optional
            The shift to apply to the annotation text. Default is [0,-0.1]
        include_intensity: bool
            If True, the intensity of the diffraction spot will be displayed with more intense peaks
            having a larger marker size.
        lazy_output: bool
            If True, the output will be a lazy signal. If None, the output will be lazy if the input is lazy.
        fast: bool
            If True the annotations will print as -h rather than \\bar{h}

        Returns
        -------
        all_markers : Sequence[hs.plot.markers.Markers]
            A list of markers for each of the n_best solutions

        """
        if lazy_output is None:
            lazy_output = self._lazy
        if text_kwargs is None:
            text_kwargs = dict()
        if annotation_shift is None:
            annotation_shift = [0, -0.15]
        if not self._lazy:
            navigation_chunks = (5,) * self.axes_manager.navigation_dimension
        else:
            navigation_chunks = None
        all_markers = []
        for n in range(n_best):
            vectors = self.to_vectors(
                n_best_index=n,
                return_object=False,
                lazy_output=True,
                navigation_chunks=navigation_chunks,
            )
            color = marker_colors[n % len(marker_colors)]
            if include_intensity:
                intensity = vectors.map(
                    vectors_to_intensity,
                    scale=intensity_scale,
                    inplace=False,
                    ragged=True,
                    output_dtype=object,
                    output_signal_size=(),
                    navigation_chunks=navigation_chunks,
                    lazy_output=True,
                    silence_warnings=True,
                ).data.T
                kwargs["sizes"] = intensity

            coords = vectors.map(
                vectors_to_coordinates,
                inplace=False,
                ragged=True,
                output_dtype=object,
                output_signal_size=(),
                navigation_chunks=navigation_chunks,
                lazy_output=True,
                silence_warnings=True,
            )
            markers = hs.plot.markers.Points.from_signal(
                coords, facecolor="none", edgecolor=color, **kwargs
            )
            all_markers.append(markers)

            if annotate:
                texts = vectors.map(
                    vectors_to_text,
                    inplace=False,
                    lazy_output=True,
                    ragged=True,
                    output_dtype=object,
                    output_signal_size=(),
                    silence_warnings=True,
                    fast=fast,
                )
                # New signal for offset coordinates, as using inplace=True shifts the point markers too
                text_coords = coords.map(
                    lambda x: x + annotation_shift,
                    inplace=False,
                    lazy_output=True,
                    silence_warnings=True,
                )
                text_markers = hs.plot.markers.Texts.from_signal(
                    text_coords, texts=texts.data.T, color=text_color, **text_kwargs
                )
                all_markers.append(text_markers)
        if lazy_output is False:  # Compute all at once (as it is faster)
            all_markers = compute_markers(all_markers)
        return all_markers

    @deprecated(
        since="0.20.0",
        alternative="pyxem.signals.OrientationMap.to_polar_markers",
        removal="1.0.0",
    )
    def to_single_phase_polar_markers(
        self,
        signal_axes: Sequence[BaseDataAxis],
        n_best: int = 1,
        marker_colors: str = ("red", "blue", "green", "orange", "purple"),
        lazy_output: bool = None,
        **kwargs,
    ) -> Iterator[hs.plot.markers.Markers]:
        return self.to_polar_markers(
            n_best=n_best,
            marker_colors=marker_colors,
            lazy_output=lazy_output,
            **kwargs,
        )

    def to_polar_markers(
        self,
        n_best: int = 1,
        marker_colors: str = ("red", "blue", "green", "orange", "purple"),
        lazy_output: bool = None,
        **kwargs,
    ) -> Iterator[hs.plot.markers.Markers]:
        """
        Convert the orientation map to a set of markers for plotting in polar coordinates.

        Parameters
        ----------
        signal_axes: Sequence[BaseDataAxis]
            The signal axes for the orientation map. The first axis should be the azimuthal axis
            and the second axis should be the radial axis.
        n_best: int
            The number of best fit solutions to return as markers
        marker_colors:
            The colors to use for the markers. If there are more than 5 solutions, the colors will repeat.
        lazy_output:
            If True, the output will be a set of lazy markers. If False the output will be a set of computed markers.
            If None, the output will be lazy if the input is lazy or not lazy if the input is not lazy.
        kwargs:
            Additional keyword arguments to pass to the hyperspy.plot.markers.Points.from_signal function.
        Returns
        -------
        all_markers : Sequence[hs.plot.markers.Markers]
            An list of markers for each of the n_best solutions

        """
        if lazy_output is None:
            lazy_output = self._lazy
        if not lazy_output:
            navigation_chunks = (5,) * self.axes_manager.navigation_dimension
        else:
            navigation_chunks = None

        def vec2polar(vector):
            # Bug with dtype of python float
            vector = vector.astype(np.float64)
            r = np.linalg.norm(vector[:, :2], axis=1)
            theta = np.arctan2(
                vector[:, 1],
                vector[:, 0],
            )
            # flip y
            theta = -theta
            return np.vstack([theta, r]).T

        all_markers = []
        for n in range(n_best):
            color = marker_colors[n % len(marker_colors)]

            vecs = self.to_vectors(
                n_best_index=n,
                return_object=False,
                lazy_output=True,
                navigation_chunks=navigation_chunks,
            )
            markers_signal = vecs.map(
                vec2polar,
                inplace=False,
                ragged=True,
                silence_warnings=True,
            )

            if "sizes" not in kwargs:
                kwargs["sizes"] = 15
            markers = hs.plot.markers.Points.from_signal(
                markers_signal, facecolor="none", edgecolor=color, **kwargs
            )

            all_markers.append(markers)
        if lazy_output is False:  # Compute all at once (as it is faster)
            all_markers = compute_markers(all_markers)
        return all_markers

    def to_ipf_colormap(
        self,
        direction: Vector3d = Vector3d.zvector(),
        add_markers: bool = True,
    ):
        """Create a colored navigator and a legend (in the form of a marker) which can be passed as the
        navigator argument to the `plot` method of some signal.

        Parameters
        ----------
        direction : Vector3d
            The direction to plot the IPF in
        add_markers : bool
            If True, the markers for the IPF will be added to the navigator as permanent markers.

        Returns
        -------
        hs.signals.BaseSignal
        """
        oris = self.to_single_phase_orientations()[:, :, 0]
        ipfcolorkey = IPFColorKeyTSL(oris.symmetry, direction)

        float_rgb = ipfcolorkey.orientation2color(oris)
        int_rgb = (float_rgb * 255).astype(np.uint8)

        s = hs.signals.Signal1D(int_rgb)
        s.change_dtype("rgb8")
        s = s.T

        if add_markers:
            annotations = get_ipf_annotation_markers(self.simulation.phases)
            s.add_marker(
                annotations,
                permanent=True,
                plot_signal=False,
                plot_marker=False,
                plot_on_signal=False,
            )
        return s

    def to_phase_map(self):
        """Create a colored navigator which can be passed as the
        navigator argument to the `plot` method of some signal.

        Returns
        -------
        hs.signals.BaseSignal
        """
        if not self.simulation.has_multiple_phases:
            raise ValueError("Only a single phase present in simulation")

        phase_idxs = self.to_phase_index()
        # in case n_best = 1
        phase_idxs = phase_idxs.reshape(
            *self.axes_manager._navigation_shape_in_array, -1
        )
        colors = [p.color_rgb for p in self.simulation.phases]

        float_rgb = np.take(colors, phase_idxs[..., 0], axis=0)
        int_rgb = (float_rgb * 255).astype(np.uint8)

        s = hs.signals.Signal1D(int_rgb)
        s.change_dtype("rgb8")
        s = s.T
        return s

    def plot_over_signal(
        self,
        signal,
        add_vector_markers=True,
        add_ipf_markers=True,
        add_ipf_correlation_heatmap=False,
        add_ipf_colorkey=True,
        vector_kwargs=None,
        **kwargs,
    ):
        """Convenience method to plot the orientation map and the n-best matches over the signal.

        Parameters
        ----------
        signal : BaseSignal
            The signal to plot the orientation map over.
        add_vector_markers : bool
            If True, the vector markers will be added to the signal.
        add_ipf_markers : bool
            If True, the IPF best fit will be added to the signal in an overlay
        add_ipf_correlation_heatmap : bool
            If True, a correlation score heatmap as an IPF will be added to the signal in an overlay.
            This overrides the `add_ipf_markers` parameter.
        add_ipf_colorkey : bool
            If True, the IPF colorkey will be added to the signal
        vector_kwargs : dict
            Additional keyword arguments to pass to the `to_single_phase_markers` method
        kwargs
            Additional keyword arguments to pass to the
            :meth:`hyperspy.api.signals.Signal2D.plot` method
        """
        if self.simulation.has_multiple_phases:
            nav = self.to_phase_map()
            add_ipf_colorkey = False
        else:
            nav = self.to_ipf_colormap(add_markers=False)
        if vector_kwargs is None:
            vector_kwargs = dict()
        signal.plot(navigator=nav, **kwargs)
        if add_vector_markers:
            signal.add_marker(self.to_markers(1, **vector_kwargs))
        if add_ipf_markers and not add_ipf_correlation_heatmap:
            ipf_markers = self.to_ipf_markers()
            signal.add_marker(ipf_markers)
        if add_ipf_correlation_heatmap:
            heatmap_markers = self.to_ipf_correlation_heatmap_markers()
            signal.add_marker(heatmap_markers)
        if add_ipf_colorkey:
            signal.add_marker(
                get_ipf_annotation_markers(self.simulation.phases), plot_on_signal=False
            )


class GenericMatchingResults:
    def __init__(self, data):
        self.data = hs.signals.Signal2D(data)

    def to_crystal_map(self):
        """
        Exports an indexation result with multiple results per navigation position to
        crystal map with one result per pixel

        Returns
        -------
        :class:`~orix.crystal_map.CrystalMap`

        """
        _s = self.data.map(
            _get_best_match,
            inplace=False,
            silence_warnings=True,
        )

        """ Gets properties """
        phase_id = _s.isig[0].data.flatten()
        alpha = _s.isig[1].data.flatten()
        beta = _s.isig[2].data.flatten()
        gamma = _s.isig[3].data.flatten()
        score = _s.isig[4].data.flatten()

        """ Gets navigation placements """
        xy = np.indices(_s.data.shape[:2])
        x = xy[1].flatten()
        y = xy[0].flatten()

        """ Tidies up so we can put these things into CrystalMap """
        euler = np.deg2rad(np.vstack((alpha, beta, gamma)).T)
        rotations = Rotation.from_euler(
            euler, convention="bunge", direction="crystal2lab"
        )

        """ add various properties """
        phase_reliabilty = self.data.map(
            _get_phase_reliability,
            inplace=False,
            silence_warnings=True,
        ).data.flatten()
        second_phase = self.data.map(
            _get_second_best_phase,
            inplace=False,
            silence_warnings=True,
        ).data.flatten()
        properties = {
            "score": score,
            "phase_reliabilty": phase_reliabilty,
            "second_phase": second_phase,
        }

        return CrystalMap(
            rotations=rotations, phase_id=phase_id, x=x, y=y, prop=properties
        )


class LazyOrientationMap(OrientationMap, LazySignal):
    pass


class VectorMatchingResults(BaseSignal):
    """Vector matching results containing the top n best matching crystal
    phase and orientation at each navigation position with associated metrics.

    Attributes
    ----------
    vectors : pyxem.signals.DiffractionVectors
        Diffraction vectors indexed.
    hkls : BaseSignal
        Miller indices associated with each diffraction vector.
    """

    _signal_dimension = 0
    _signal_type = "vector_matching"

    def __init__(self, *args, **kwargs):
        BaseSignal.__init__(self, *args, **kwargs)
        # self.axes_manager.set_signal_dimension(2)
        self.vectors = None
        self.hkls = None

    def get_crystallographic_map(self, *args, **kwargs):
        """Obtain a crystallographic map specifying the best matching phase and
        orientation at each probe position with corresponding metrics.

        Returns
        -------
        cryst_map : Signal2D
            Crystallographic mapping results containing the best matching phase
            and orientation at each navigation position with associated metrics.
            The Signal at each navigation position is an array of,
            [phase, np.array((z,x,z)), dict(metrics)]
            which defines the phase, orientation as Euler angles in the zxz
            convention and metrics associated with the matching.
            Metrics for template matching results are
            'match_rate'
            'total_error'
            'orientation_reliability'
            'phase_reliability'
        """
        crystal_map = self.map(
            crystal_from_vector_matching,
            inplace=False,
            silence_warnings=True,
            *args,
            **kwargs,
        )

        crystal_map = _transfer_navigation_axes(crystal_map, self)
        return crystal_map

    def get_indexed_diffraction_vectors(
        self, vectors, overwrite=False, *args, **kwargs
    ):
        """Obtain an indexed diffraction vectors object.

        Parameters
        ----------
        vectors : pyxem.signals.DiffractionVectors
            A diffraction vectors object to be indexed.

        Returns
        -------
        indexed_vectors : pyxem.signals.DiffractionVectors
            An indexed diffraction vectors object.

        """
        if overwrite is False:
            if vectors.hkls is not None:
                warn(
                    "The vectors supplied are already associated with hkls set "
                    "overwrite=True to replace these hkls."
                )
            else:
                vectors.hkls = self.hkls

        elif overwrite is True:
            vectors.hkls = self.hkls
        return vectors



# - * - coding: utf - 8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from pyxem.signals.diffraction_vectors2d import DiffractionVectors2D
import numpy as np
import matplotlib.pyplot as plt


import hyperspy.api as hs

from pyxem.utils.vectors import (
    column_mean,
    vectors2image,
    convert_to_markers,
    points_to_polygon,
)
from pyxem.utils.vectors import only_signal_axes


class LabeledDiffractionVectors2D(DiffractionVectors2D):
    """The labeled diffraction vectors 2d signal"""

    _signal_dimension = 2
    _signal_type = "labeled_diffraction_vectors"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_clustered = False

    @property
    def is_clustered(self):
        return self.metadata.VectorMetadata["is_clustered"]

    @is_clustered.setter
    def is_clustered(self, value):
        self.metadata.VectorMetadata["is_clustered"] = value

    @only_signal_axes
    def map_vectors(self, func, dtype, label_index=-1, shape=None, **kwargs):
        """
        Parameters
        ----------
        func: func
            The function to be applied to each label
        label_index: int
            The index of the label to be analyzed. Usually this is the last column
        dtype: np.dtype
            The dtype used to initialize the output array
        shape: None or tuple
            The shape for the output of the `func`
        kwargs:
            Any additional arguments passed to `func`
        Returns
        -------
        signal:
            The signal from the map_vectors function. Each index on the navigation axes
            is related to the label for each vector.
        """
        vectors = self.data
        labels = vectors[:, label_index]
        label_order = labels.argsort()
        labels = labels[label_order]  # Order the labels
        vectors = vectors[label_order]  # Order the vectors

        sorted_index = np.arange(0, np.max(labels) + 1)
        lo = np.searchsorted(labels, sorted_index, side="left")
        hi = np.searchsorted(labels, sorted_index, side="right")

        if shape is not None:
            ans = np.empty((len(sorted_index),) + shape, dtype=dtype)
        else:
            ans = np.empty(len(sorted_index), dtype=dtype)
        for i, (l, h) in enumerate(zip(lo, hi)):
            ans[i] = func(vectors[l:h], **kwargs)
        return ans

    @only_signal_axes
    def plot_clustered(
        self,
        nav_columms=None,
        signal_columns=None,
        navigation_pixels=(105, 105),
        scales=None,
        offsets=None,
        labels=None,
        signal=None,
        figsize=None,
    ):
        """Plot the clustered vectors and the characteristic diffraction patterns

        Parameters
        ----------
        nav_columms: None or list
            The columns to be used for the navigation axes
        signal_columns: None or list
            The columns to be used for the signal axes
        navigation_pixels: tuple
            The number of pixels in the navigation axes
        scales: None or tuple
            The scales for the navigation axes
        offsets: None or tuple
            The offsets for the navigation axes
        labels: None or list
            The labels to be plotted
        signal: None or pyxem.signals.ElectronDiffraction2D
            The signal to be plotted on the
        figsize: None or tuple
            The figure size to be used
        """
        if signal_columns is None:
            signal_columns = [2, 3]
        if nav_columms is None:
            nav_columms = [0, 1]
        if scales is None:
            nav = self.data[:, nav_columms]
            xscale = (np.max(nav[:, 0]) - np.min(nav[:, 0])) / (
                navigation_pixels[0] - 1
            )
            yscale = (np.max(nav[:, 1]) - np.min(nav[:, 1])) / (
                navigation_pixels[1] - 1
            )
            scales = (xscale, yscale)
            print(scales)
        if offsets is None:
            nav = self.data[:, nav_columms]
            offsets = (np.min(nav[:, 0]), np.min(nav[:, 1]))
        if not self.is_clustered:
            raise ValueError(
                "You must first cluster the dataset using the "
                "`cluster_labeled_vectors` function."
            )
        num_clusters = int(np.max(self.data[:, -1]) + 1)
        if num_clusters == 0:
            raise ValueError("No clusters found!")
        fig, axs = plt.subplots(2, num_clusters, figsize=figsize)
        if labels is None:
            labels = range(num_clusters)
        for i in labels:
            clustered_peaks = self.data[self.data[:, -1] == i]
            unique_peaks = np.unique(clustered_peaks[:, -2])
            for p in unique_peaks:
                is_p = clustered_peaks[:, -2] == p
                pks = clustered_peaks[is_p]
                img = vectors2image(
                    pks,
                    image_size=navigation_pixels,
                    scales=scales,
                    offsets=offsets,
                    indexes=nav_columms,
                )
                img[img == 0] = np.nan
                ext = tuple((np.array(scales) * navigation_pixels) + offsets)
                ext = (offsets[0], ext[0], offsets[1], ext[1])
                axs[0, i].imshow(img, alpha=0.2, extent=ext)
            if signal is not None:
                axs[1, i].imshow(signal.data, extent=signal.axes_manager.signal_extent)
            axs[1, i].scatter(
                clustered_peaks[:, signal_columns[1]],
                clustered_peaks[:, signal_columns[0]],
            )
        return fig, axs

    @only_signal_axes
    def cluster_labeled_vectors(
        self, method, columns=None, preprocessing="mean", **kwargs
    ):
        """A function to cluster the labeled vectors in the dataset.

        Parameters
        ----------
        method: sklearn.base.ClusterMixin
            The clustering method to be used. This is a class that implements the ``fit`` method
        columns: None or list
            The columns to be used for clustering. If None, the first two columns are used
        preprocessing: str or callable
            The function to be applied to each label clustering. If 'mean', the mean of the
            vectors is used. If callable, the function is applied to each label and the result
            is used for clustering.
        """
        if columns is None:
            columns = [0, 1]
        if preprocessing == "mean":
            preprocessing = column_mean
            kwargs = {"label_index": -1, "dtype": float, "shape": (2,)}
        elif callable(preprocessing):
            preprocessing = preprocessing
        else:
            raise ValueError("The preprocessing must be either 'mean' or a function")
        to_cluster_vectors = self.map_vectors(
            preprocessing,
            columns=columns,
            **kwargs,
        )
        # Remove the nan values but we need to keep track of the original labels
        not_nan = np.logical_not(np.any(np.isnan(to_cluster_vectors), axis=1))
        num_labels = len(to_cluster_vectors)
        to_cluster_vectors = to_cluster_vectors[not_nan]
        clustering = method.fit(to_cluster_vectors)
        labels = clustering.labels_
        initial_labels = self.data[:, -1].astype(int)
        # Replace the values with the original
        actual_labels = np.full(num_labels, -1)
        actual_labels[not_nan] = labels
        new_labels = actual_labels[initial_labels]
        new_labels[initial_labels == -1] = -1
        print(f"{np.max(labels) + 1} : Clusters Found!")
        vectors_and_labels = np.hstack([self.data, new_labels[:, np.newaxis]])
        new_signal = self._deepcopy_with_new_data(data=vectors_and_labels)
        new_signal.ragged = False  # Need to reset this (Remove once hyperspy checks for object --> ragged)
        new_signal.axes_manager.signal_axes[0].size = (
            new_signal.axes_manager.signal_axes[0].size + 1
        )
        new_signal.is_clustered = True
        new_signal.column_names = np.append(self.column_names, ["cluster_label"])
        new_signal.units = np.append(self.units, ["n.a."])
        return new_signal

    @only_signal_axes
    def to_markers(self, signal, get_polygons=False, num_points=10, **kwargs):
        """Convert the labeled vectors to markers

        Parameters
        ----------
        signal: pyxem.signals.ElectronDiffraction2D
            The signal which the markers will be plotted on
        get_polygons: bool
            If True, both the vectors and the polygons will be returned
        num_points: int
            The number of points to be used to create the polygon
        kwargs:
            Any additional arguments to be passed to the :class:`~hyperspy.api.plot.Points` Maker

        Returns
        -------
        points:
            The points to be plotted
        polygons:
            The polygons to be plotted. Only returned if `get_polygons` is True

        """
        offsets, colors, colors_by_index = convert_to_markers(self, signal)
        points = hs.plot.markers.Points(offsets=offsets.T, edgecolor=colors.T, **kwargs)
        if get_polygons:
            verts = self.map_vectors(
                points_to_polygon, num_points=num_points, dtype=object
            )
            verts = list(verts)
            polygons = hs.plot.markers.Polygons(
                verts=verts, alpha=0.5, color=colors_by_index, linewidth=2
            )
            return points, polygons
        return points



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy.signals import Signal2D
from hyperspy._signals.lazy import LazySignal

import numpy as np
from hyperspy.drawing.utils import plot_images


class Power2D(Signal2D):
    """Signal class for the power spectrum of some Correlation


    Parameters
    ----------
    *args
        See :class:`hyperspy.api.signals.Signal2D`.
    **kwargs
        See :class:`hyperspy.api.signals.Signal2D`
    """

    _signal_type = "power"

    def get_map(self, k_region=None, symmetry=None):
        """Creates a 2 dimensional map of from the power spectrum.

        Parameters
        ----------
        k_region: array-like
           upper and lower k values to integrate over, allows both ints and floats for indexing
        symmetry: int or array-like
            specific integers or list of symmetries to average over when creating the map of the correlations.

        Returns
        -------
        symmetry_map: 2-d array
            2 dimensional map of from the power spectrum
        """
        if k_region is None:
            k_region = [0, -1]
        if symmetry is None:
            sym_map = (
                self.isig[:, k_region[0] : k_region[1]].sum(axis=[-1, -2]).transpose()
            )

        elif isinstance(symmetry, int):
            sym_map = (
                self.isig[symmetry, k_region[0] : k_region[1]]
                .sum(axis=[-1])
                .transpose()
            )

        else:
            sym_map = Signal2D(data=np.zeros(self.axes_manager.navigation_shape))
            for sym in symmetry:
                sym_map = (
                    self.isig[sym, k_region[0] : k_region[1]].sum(axis=[-1]).transpose()
                    + sym_map
                )
        return sym_map

    def plot_symmetries(self, k_region=None, symmetry=None, *args, **kwargs):
        """Plots the symmetries in the list of symmetries. Plot symmetries takes all of the arguments that imshow does.

        Parameters
        ----------
         k_region: array-like
           upper and lower k values to integrate over, allows both ints and floats for indexing
        symmetry: list or None
            specific integers or list of symmetries to average over when creating the map of the correlations.
            If None, defaults to [2, 4, 6, 8, 10]
        """

        symmetry = [2, 4, 6, 8, 10] if symmetry is None else symmetry

        if k_region is None:
            k_region = [0, -1]
        summed = [self.get_map(k_region=k_region)]
        maps = summed + [self.get_map(k_region=k_region, symmetry=i) for i in symmetry]
        labels = ["summed"] + [str(i) + "-fold" for i in symmetry]
        plot_images(images=maps, label=labels, *args, **kwargs)


class LazyPower2D(LazySignal, Power2D):
    pass



import numpy as np
from orix.crystal_map import Phase
from orix.quaternion import Rotation, Orientation
from diffpy.structure import Atom, Lattice, Structure
from diffsims.generators.simulation_generator import SimulationGenerator
from scipy.ndimage import gaussian_filter1d
from pyxem.signals import Diffraction1D, Diffraction2D
import skimage


def si_phase():
    """Create a silicon phase with space group 227. This is a diamond cubic structure with
    a lattice parameter of 5.431 Å.

    """
    a = 5.431
    latt = Lattice(a, a, a, 90, 90, 90)
    atom_list = []
    for coords in [[0, 0, 0], [0.5, 0, 0.5], [0, 0.5, 0.5], [0.5, 0.5, 0]]:
        x, y, z = coords[0], coords[1], coords[2]
        atom_list.append(Atom(atype="Si", xyz=[x, y, z], lattice=latt))  # Motif part A
        atom_list.append(
            Atom(atype="Si", xyz=[x + 0.25, y + 0.25, z + 0.25], lattice=latt)
        )  # Motif part B
    struct = Structure(atoms=atom_list, lattice=latt)
    p = Phase(structure=struct, space_group=227)
    return p


def si_tilt():
    p = si_phase()
    gen = SimulationGenerator()
    rotations = Rotation.from_euler(
        [[0, 0, 0], [10, 0, 0]],
        degrees=True,
    )
    sim = gen.calculate_diffraction2d(
        phase=p, rotation=rotations, reciprocal_radius=1.5, max_excitation_error=0.1
    )
    dp1 = np.flipud(
        sim.get_diffraction_pattern(sigma=5, shape=(128, 128))
    )  # flip up/down to go from scatter to diffraction
    dp2 = np.flipud(sim.irot[1].get_diffraction_pattern(sigma=5, shape=(128, 128)))

    top = np.tile(
        dp1,
        (4, 2, 1, 1),
    )
    bottom = np.tile(
        dp2,
        (4, 2, 1, 1),
    )
    total = np.hstack((top, bottom))
    tilt = Diffraction2D(total)
    tilt.axes_manager.signal_axes[0].name = "kx"
    tilt.axes_manager.signal_axes[1].name = "kx"
    tilt.axes_manager.signal_axes[0].units = r"$\AA^{-1}$"
    tilt.axes_manager.signal_axes[1].units = r"$\AA^{-1}$"
    tilt.axes_manager.signal_axes[0].scale = 0.01
    tilt.axes_manager.signal_axes[1].scale = 0.01
    return tilt


def si_grains_from_orientations(
    oris: Orientation, seed: int = 2, size: int = 20, recip_pixels: int = 128
):
    p = si_phase()
    gen = SimulationGenerator()
    num_grains = oris.size

    sim = gen.calculate_diffraction2d(
        phase=p, rotation=oris, reciprocal_radius=1.5, max_excitation_error=0.1
    )
    dps = [
        np.flipud(
            sim.irot[i].get_diffraction_pattern(
                sigma=5, shape=(recip_pixels, recip_pixels)
            )
        )
        for i in range(num_grains)
    ]
    rng = np.random.default_rng(seed)
    x = rng.integers(0, size, size=num_grains)
    y = rng.integers(0, size, size=num_grains)
    navigator = np.zeros((size, size))
    for i in range(num_grains):
        navigator[x[i], y[i]] = i + 1
    navigator = skimage.segmentation.expand_labels(navigator, distance=size)
    grain_data = np.empty((size, size, recip_pixels, recip_pixels))
    for i in range(num_grains):
        grain_data[navigator == i + 1] = dps[i][np.newaxis]
    grains = Diffraction2D(grain_data)
    grains.axes_manager.signal_axes[0].name = "kx"
    grains.axes_manager.signal_axes[1].name = "kx"
    grains.axes_manager.signal_axes[0].units = r"$\AA^{-1}$"
    grains.axes_manager.signal_axes[1].units = r"$\AA^{-1}$"
    grains.axes_manager.signal_axes[0].scale = 0.01
    grains.axes_manager.signal_axes[1].scale = 0.01
    return grains


def si_grains(num_grains=4, seed=2, size=6, recip_pixels=128, return_rotations=False):
    """Generate a simulated dataset with grains in random orientations"""
    p = si_phase()
    rotations = Orientation.random(num_grains, symmetry=p.point_group)
    grains = si_grains_from_orientations(
        rotations, seed=seed, size=size, recip_pixels=recip_pixels
    )

    if return_rotations:
        return grains, rotations
    else:
        return grains


def si_grains_simple(seed=2, size=6, recip_pixels=128, return_rotations=False):
    """Generate a simulated dataset with low-index zone axes"""
    p = si_phase()
    rotations = Orientation.from_euler(
        [
            [0, 0, 0],  # [0 0 1]
            [0, 45, 0],  # [0 1 1]
            [0, 54.7, 45],  # [1 1 1]
            [0, 35, 45],  # [1 1 2]
        ],
        degrees=True,
        symmetry=p.point_group,
    )
    grains = si_grains_from_orientations(
        rotations, seed=seed, size=size, recip_pixels=recip_pixels
    )

    if return_rotations:
        return grains, rotations
    else:
        return grains


def si_rotations_line():
    from orix.sampling import get_sample_reduced_fundamental

    p = si_phase()
    gen = SimulationGenerator()
    rotations = get_sample_reduced_fundamental(resolution=3, point_group=p.point_group)
    sim = gen.calculate_diffraction2d(
        phase=p, rotation=rotations, max_excitation_error=0.1, reciprocal_radius=2
    )
    dps = []
    for i in range(rotations.size):
        dp = np.flipud(sim.irot[i].get_diffraction_pattern(sigma=5, shape=(256, 256)))
        dps.append(dp)
    line = Diffraction2D(np.array(dps))
    line.axes_manager.signal_axes[0].name = "kx"
    line.axes_manager.signal_axes[1].name = "kx"
    line.axes_manager.signal_axes[0].units = r"$\AA^{-1}$"
    line.axes_manager.signal_axes[1].units = r"$\AA^{-1}$"
    line.axes_manager.signal_axes[0].scale = 0.01
    line.axes_manager.signal_axes[1].scale = 0.01
    return line


def simulated1dsi(
    num_points=200,
    accelerating_voltage=200,
    reciporical_radius=1,
    sigma=2,
):
    p = si_phase()
    gen = SimulationGenerator(
        accelerating_voltage=accelerating_voltage,
    )
    x = np.zeros(num_points)
    sim = gen.calculate_diffraction1d(phase=p, reciprocal_radius=reciporical_radius)
    int_points = np.round(
        np.array(sim.reciprocal_spacing) * num_points / reciporical_radius
    ).astype(int)
    x[int_points] = sim.intensities
    x = gaussian_filter1d(x, sigma)
    s = Diffraction1D(x)
    s.axes_manager[0].name = "k"
    s.axes_manager[0].units = r"$\AA^{-1}$"
    s.axes_manager[0].scale = reciporical_radius / num_points
    return s



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from pytest import approx
import numpy as np
from numpy.testing import assert_allclose

import hyperspy.api as hs
import pyxem.utils._beam_shift_tools as bst


class TestGetRgbPhaseMagnitudeArray:
    def test_simple(self):
        phase = np.zeros((50, 50))
        magnitude = np.zeros((50, 50))
        rgb_array = bst._get_rgb_phase_magnitude_array(phase, magnitude)
        assert (rgb_array == 0.0).all()

    def test_magnitude_zero(self):
        phase = np.random.random((50, 50))
        magnitude = np.zeros((50, 50))
        rgb_array = bst._get_rgb_phase_magnitude_array(phase, magnitude)
        assert (rgb_array == 0.0).all()

    def test_all_same(self):
        phase = np.ones((50, 50))
        magnitude = np.ones((50, 50))
        rgb_array = bst._get_rgb_phase_magnitude_array(phase, magnitude)
        assert (rgb_array == rgb_array[0][0]).all()


class TestGetRgbPhaseArray:
    def test_all_same0(self):
        phase = np.zeros((50, 50))
        rgb_array = bst._get_rgb_phase_array(phase)
        assert (rgb_array == rgb_array[0][0]).all()

    def test_all_same1(self):
        phase = np.ones((50, 50))
        rgb_array = bst._get_rgb_phase_array(phase)
        assert (rgb_array == rgb_array[0][0]).all()


class TestFindPhase:
    def test_simple(self):
        phase = np.zeros((50, 50))
        phase0 = bst._find_phase(phase)
        assert (phase0 == 0.0).all()

    def test_rotation(self):
        phase = np.zeros((50, 50))
        phase0 = bst._find_phase(phase, rotation=90)
        assert (phase0 == np.pi / 2).all()
        phase1 = bst._find_phase(phase, rotation=45)
        assert (phase1 == np.pi / 4).all()
        phase2 = bst._find_phase(phase, rotation=180)
        assert (phase2 == np.pi).all()
        phase3 = bst._find_phase(phase, rotation=360)
        assert (phase3 == 0).all()
        phase4 = bst._find_phase(phase, rotation=-90)
        assert (phase4 == 3 * np.pi / 2).all()

    def test_max_phase(self):
        phase = (np.ones((50, 50)) * np.pi * 0.5) + np.pi
        phase0 = bst._find_phase(phase, max_phase=np.pi)
        assert (phase0 == np.pi / 2).all()


class TestGetCornerSlices:
    @pytest.mark.parametrize("corner_size", [0.02, 0.05, 0.20, 0.23])
    def test_corner_size(self, corner_size):
        size = 100
        s = hs.signals.Signal2D(np.zeros((size, size))).T
        corner_slice_list = bst._get_corner_slices(s, corner_size=corner_size)
        corner_shape = (round(size * corner_size), round(size * corner_size))
        for corner_slice in corner_slice_list:
            s_corner = s.inav[corner_slice]
            assert s_corner.axes_manager.shape == corner_shape

    def test_signal_slice_values(self):
        data = np.zeros((100, 200), dtype=np.uint16)
        data[:20, :20] = 2
        data[:20, -20:] = 5
        data[-20:, :20] = 10
        data[-20:, -20:] = 8
        s = hs.signals.Signal2D(data).T
        corner_slice_list = bst._get_corner_slices(s, corner_size=0.05)
        assert (s.inav[corner_slice_list[0]].data == 2).all()
        assert (s.inav[corner_slice_list[1]].data == 10).all()
        assert (s.inav[corner_slice_list[2]].data == 5).all()
        assert (s.inav[corner_slice_list[3]].data == 8).all()

    def test_non_square_signal(self):
        corner_size = 0.05
        size_x, size_y = 100, 200
        s = hs.signals.Signal2D(np.zeros((size_y, size_x))).T
        corner_slice_list = bst._get_corner_slices(s, corner_size=0.05)
        corner_shape = (round(size_x * corner_size), round(size_y * corner_size))
        for corner_slice in corner_slice_list:
            s_corner = s.inav[corner_slice]
            assert s_corner.axes_manager.shape == corner_shape

    def test_wrong_input_dimensions(self):
        s = hs.signals.Signal2D(np.ones((2, 10, 10)))
        with pytest.raises(ValueError):
            bst._get_corner_slices(s)
        s = hs.signals.Signal2D(np.ones((2, 2, 10, 10)))
        with pytest.raises(ValueError):
            bst._get_corner_slices(s)
        s = hs.signals.Signal1D(np.ones(10))
        with pytest.raises(ValueError):
            bst._get_corner_slices(s)


class TestPlaneParametersToImage:
    def test_simple(self):
        p = [0, 0, 1, 0]
        xaxis, yaxis = range(100), range(110)
        image = bst._plane_parameters_to_image(p, xaxis, yaxis)
        assert_allclose(image, 0)

    def test_offset(self):
        p = [0, 0, 1, 3]
        xaxis, yaxis = range(100), range(110)
        image = bst._plane_parameters_to_image(p, xaxis, yaxis)
        assert_allclose(image, -3)

    def test_x_plane(self):
        p = [1, 0, 1, 0]
        xaxis, yaxis = range(100), range(110)
        image = bst._plane_parameters_to_image(p, xaxis, yaxis)
        assert image[0, 0] > image[0, -1]

    def test_y_plane(self):
        p = [0, 1, 1, 0]
        xaxis, yaxis = range(100), range(110)
        image = bst._plane_parameters_to_image(p, xaxis, yaxis)
        assert image[0, 0] > image[-1, 0]

    def test_last_parameter(self):
        p = [0, 0, 2, 4]
        xaxis, yaxis = range(100), range(110)
        image = bst._plane_parameters_to_image(p, xaxis, yaxis)
        assert_allclose(image, -2)


class TestGetLinearPlaneFromSignal2d:
    def test_linear_ramp(self):
        s0, s1 = hs.signals.Signal2D(np.meshgrid(range(100), range(110)))
        s0.change_dtype("float64")
        s1.change_dtype("float64")
        s0_plane = bst._get_linear_plane_from_signal2d(s0)
        s1_plane = bst._get_linear_plane_from_signal2d(s1)
        np.testing.assert_almost_equal(s0_plane.data, s0.data)
        np.testing.assert_almost_equal(s1_plane.data, s1.data)

    def test_zeros_values(self):
        s = hs.signals.Signal2D(np.zeros((100, 200), dtype=np.float32))
        s_plane = bst._get_linear_plane_from_signal2d(s)
        np.testing.assert_almost_equal(s_plane.data, s.data)

    def test_ones_values(self):
        s = hs.signals.Signal2D(np.ones((100, 200), dtype=np.float32))
        s_plane = bst._get_linear_plane_from_signal2d(s)
        np.testing.assert_almost_equal(s_plane.data, s.data)

    def test_negative_values(self):
        data = np.ones((110, 100)) * -10
        s = hs.signals.Signal2D(data)
        s_plane = bst._get_linear_plane_from_signal2d(s)
        np.testing.assert_almost_equal(s_plane.data, s.data)

    def test_mask(self):
        data = np.ones((110, 200), dtype=np.float32)
        mask = np.zeros_like(data, dtype=bool)
        data[50, 51] = 10000
        mask[50, 51] = True
        s = hs.signals.Signal2D(data)
        plane_no_mask = bst._get_linear_plane_from_signal2d(s)
        plane_mask = bst._get_linear_plane_from_signal2d(s, mask=mask)
        assert plane_no_mask != approx(1.0)
        assert plane_mask == approx(1.0)

    def test_offest_and_scale(self):
        s, _ = hs.signals.Signal2D(np.meshgrid(range(100), range(110)))
        s.axes_manager[0].offset = -40
        s.axes_manager[1].offset = 50
        s.axes_manager[0].scale = -0.22
        s.axes_manager[1].scale = 5.2
        s_orig = s.deepcopy()
        mask = np.zeros_like(s.data, dtype=bool)
        s.data[50, 51] = 10000
        mask[50, 51] = True
        plane_mask = bst._get_linear_plane_from_signal2d(s, mask=mask)
        np.testing.assert_allclose(plane_mask, s_orig.data, atol=1e-6)

    def test_crop_signal(self):
        s, _ = hs.signals.Signal2D(np.meshgrid(range(100), range(110)))
        s.change_dtype("float32")
        s.axes_manager[0].offset = -40
        s.axes_manager[1].offset = 50
        s.axes_manager[0].scale = -0.22
        s.axes_manager[1].scale = 5.2
        s_crop = s.isig[10:-20, 21:-11]
        s_crop_orig = s_crop.deepcopy()
        s_crop.data[50, 51] = 10000
        mask = np.zeros_like(s_crop.data, dtype=bool)
        mask[50, 51] = True
        plane = bst._get_linear_plane_from_signal2d(s_crop, mask=mask)
        np.testing.assert_almost_equal(plane, s_crop_orig.data, decimal=6)

    def test_wrong_input_dimensions(self):
        s = hs.signals.Signal2D(np.ones((2, 10, 10)))
        with pytest.raises(ValueError):
            bst._get_linear_plane_from_signal2d(s)
        s = hs.signals.Signal2D(np.ones((2, 2, 10, 10)))
        with pytest.raises(ValueError):
            bst._get_linear_plane_from_signal2d(s)
        s = hs.signals.Signal1D(np.ones(10))
        with pytest.raises(ValueError):
            bst._get_linear_plane_from_signal2d(s)

    def test_wrong_mask_dimensions(self):
        s = hs.signals.Signal2D(np.ones((10, 10)))
        mask = np.zeros((11, 9), dtype=bool)
        with pytest.raises(ValueError):
            bst._get_linear_plane_from_signal2d(s, mask=mask)


class TestGetLimitsFromArray:
    def test_simple(self):
        data_array0 = np.array((5, 10))
        clim0 = bst._get_limits_from_array(data_array0, sigma=4)
        assert (data_array0.min(), data_array0.max()) == clim0
        data_array1 = np.array((5, -10))
        clim1 = bst._get_limits_from_array(data_array1, sigma=4)
        assert (data_array1.min(), data_array1.max()) == clim1
        data_array2 = np.array((-5, 10))
        clim2 = bst._get_limits_from_array(data_array2, sigma=4)
        assert (data_array2.min(), data_array2.max()) == clim2
        data_array3 = np.array((-5, -10))
        clim3 = bst._get_limits_from_array(data_array3, sigma=4)
        assert (data_array3.min(), data_array3.max()) == clim3

    def test_simple_sigma(self):
        data_array0 = np.array((5, 10))
        clim0 = bst._get_limits_from_array(data_array0, sigma=0)
        assert (data_array0.mean(), data_array0.mean()) == clim0
        data_array1 = np.array((5, -10))
        clim1 = bst._get_limits_from_array(data_array1, sigma=0)
        assert (data_array1.mean(), data_array1.mean()) == clim1
        data_array2 = np.array((-5, 10))
        clim2 = bst._get_limits_from_array(data_array2, sigma=0)
        assert (data_array2.mean(), data_array2.mean()) == clim2
        data_array3 = np.array((-5, -10))
        clim3 = bst._get_limits_from_array(data_array3, sigma=0)
        assert (data_array3.mean(), data_array3.mean()) == clim3

    def test_ignore_zeros(self):
        data_array0 = np.zeros(shape=(100, 100))
        value = 50
        data_array0[:, 70:80] = value
        clim0_0 = bst._get_limits_from_array(data_array0, ignore_zeros=True)
        assert (value, value) == clim0_0
        clim0_1 = bst._get_limits_from_array(data_array0, sigma=0)
        assert (data_array0.mean(), data_array0.mean()) == clim0_1
        clim0_2 = bst._get_limits_from_array(data_array0, sigma=1)
        assert (0.0, 20.0) == clim0_2

    def test_ignore_edges(self):
        data_array = np.ones(shape=(100, 100)) * 5000
        value = 50
        data_array[1:-1, 1:-1] = value
        clim0 = bst._get_limits_from_array(data_array, ignore_edges=True)
        assert (value, value) == clim0
        clim1 = bst._get_limits_from_array(data_array, ignore_edges=False)
        assert not ((value, value) == clim1)


class TestMakeBivariateHistogram:
    def test_make_bivariate_histogram(self):
        x, y = np.ones((100, 100)), np.ones((100, 100))
        bst._make_bivariate_histogram(
            x_position=x,
            y_position=y,
            histogram_range=None,
            masked=None,
            bins=200,
            spatial_std=3,
        )

    def test_single_x(self):
        size = 100
        x, y = np.ones(size), np.zeros(size)
        s = bst._make_bivariate_histogram(x, y)
        hist_iX = s.axes_manager[0].value2index(1.0)
        hist_iY = s.axes_manager[1].value2index(0.0)
        assert s.data[hist_iY, hist_iX] == size
        s.data[hist_iY, hist_iX] = 0
        assert not s.data.any()

    def test_single_negative_x(self):
        size = 100
        x, y = -np.ones(size), np.zeros(size)
        s = bst._make_bivariate_histogram(x, y)
        hist_iX = s.axes_manager[0].value2index(-1)
        hist_iY = s.axes_manager[1].value2index(0)
        assert s.data[hist_iY, hist_iX] == size
        s.data[hist_iY, hist_iX] = 0
        assert not s.data.any()

    def test_single_negative_x_y(self):
        size = 100
        x, y = -np.ones(size), np.ones(size)
        s = bst._make_bivariate_histogram(x, y)
        hist_iX = s.axes_manager[0].value2index(-1)
        hist_iY = s.axes_manager[1].value2index(1)
        assert s.data[hist_iY, hist_iX] == size
        s.data[hist_iY, hist_iX] = 0
        assert not s.data.any()



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from hyperspy.axes import UniformDataAxis

from pyxem.utils import calibration
from pyxem.utils.calibration import Calibration
from pyxem.signals import Diffraction2D


class TestCalibrationClass:
    @pytest.fixture
    def calibration(self):
        s = Diffraction2D(np.zeros((10, 10)))
        return Calibration(s)

    def test_init(self, calibration):
        assert isinstance(calibration, Calibration)

    def test_set_center(self, calibration):
        calibration(center=(5, 5))
        assert calibration.signal.axes_manager[0].offset == -5
        assert calibration.signal.axes_manager[1].offset == -5
        assert calibration.flat_ewald is True

    def test_set_units(self, calibration):
        calibration(units="k_nm^-1")
        assert calibration.signal.axes_manager[0].units == "k_nm^-1"
        assert calibration.signal.axes_manager[1].units == "k_nm^-1"
        assert calibration.units == ["k_nm^-1", "k_nm^-1"]

    def test_calibrate_gain(self, calibration):
        calibration.signal.data[0, 0] = 1
        calibration.detector_gain = 2
        assert calibration.detector_gain == 2
        assert calibration.signal.data[0, 0] == 0.5

    def test_set_mask(self, calibration):
        calibration(mask=np.ones((10, 10)))
        assert calibration.mask is not None
        assert calibration.mask.shape == (10, 10)

    def test_set_affine(self, calibration):
        assert calibration.affine is None
        calibration(affine=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
        np.testing.assert_array_equal(
            calibration.affine, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        )

    def test_set_beam_energy(self, calibration):
        assert calibration.beam_energy is None
        calibration(beam_energy=200)
        assert calibration.beam_energy == 200
        assert calibration.wavelength is not None

    def test_set_wavelength(self, calibration):
        assert calibration.wavelength is None
        calibration(wavelength=0.02508)
        assert calibration.wavelength == 0.02508

    def test_get_camera_length(self, calibration):
        assert calibration.camera_length is None
        calibration.camera_length = 0.038
        assert calibration.camera_length == 0.038

    def test_set_convergence_angle(self, calibration):
        assert calibration.convergence_angle is None
        calibration.convergence_angle = 1.5
        assert calibration.convergence_angle == 1.5

    def test_set_scale(self, calibration):
        calibration(scale=0.01)
        assert calibration.signal.axes_manager[0].scale == 0.01
        assert calibration.signal.axes_manager[1].scale == 0.01
        assert calibration.scale == [0.01, 0.01]
        assert calibration.flat_ewald is True

    def test_set_camera_length_from_calibration(self, calibration):
        calibration.scale = 0.156
        calibration.units = "nm^-1"
        calibration.beam_energy = 200
        calibration.physical_pixel_size = 15e-6  # 15 um
        calibration.set_camera_length_from_calibration()
        assert calibration.camera_length == pytest.approx(0.038, rel=1e-2)

    @pytest.mark.parametrize("units", ["nm^-1", "A^-1", "px", "mrad"])
    def test_change_signal_units(self, calibration, units):
        calibration.scale = 0.156
        calibration.units = "nm^-1"
        calibration.beam_energy = 200
        calibration.physical_pixel_size = 15e-6  # 15 um
        calibration.convert_signal_units(units)
        if units == "nm^-1":
            np.testing.assert_almost_equal(calibration.scale, [0.156, 0.156])
        elif units == "A^-1":
            np.testing.assert_almost_equal(calibration.scale, [0.0156, 0.0156])
        elif units == "px":
            np.testing.assert_almost_equal(calibration.scale, [1, 1])
        calibration.convert_signal_units("nm^-1")
        np.testing.assert_almost_equal(calibration.scale, [0.156, 0.156])

    @pytest.mark.parametrize("units", ["nm^-1", "A^-1", "px", "mrad"])
    def test_change_signal_units_rebin(self, calibration, units):
        s = calibration.signal
        s.calibration.scale = 0.156
        s.calibration.units = "nm^-1"
        s.calibration.beam_energy = 200
        s.calibration.physical_pixel_size = 15e-6  # 15 um

        s_rebin = s.calibration.signal.rebin(scale=(2, 1))
        assert s_rebin.axes_manager[0].scale == 0.156 * 2
        np.testing.assert_almost_equal(s_rebin.calibration.scale, [0.156 * 2, 0.156])
        s_rebin.calibration.convert_signal_units(units)
        if units == "nm^-1":
            np.testing.assert_almost_equal(
                s_rebin.calibration.scale, [0.156 * 2, 0.156]
            )
        elif units == "A^-1":
            np.testing.assert_almost_equal(
                s_rebin.calibration.scale, [0.0156 * 2, 0.0156]
            )
        elif units == "px":
            np.testing.assert_almost_equal(s_rebin.calibration.scale, [1, 1])
        s_rebin.calibration.convert_signal_units("nm^-1")
        np.testing.assert_almost_equal(s_rebin.calibration.scale, [0.156 * 2, 0.156])

    def test_set_failure(self, calibration):
        assert calibration.wavelength is None
        assert calibration.beam_energy is None
        with pytest.raises(ValueError):
            calibration.detector(pixel_size=0.1, detector_distance=1)
        calibration.beam_energy = 200
        calibration.detector(pixel_size=0.1, detector_distance=1)
        # The center, in pixel coordinates, is (4.5, 4.5)
        # When using a detector, this gets rounded down to 4
        assert calibration.center == [4, 4]
        calibration.detector(
            pixel_size=0.1, detector_distance=1, beam_energy=200, units="k_nm^-1"
        )
        assert calibration.flat_ewald is False
        assert calibration.center == [4, 4]
        with pytest.raises(ValueError):
            calibration(scale=0.01)
        assert calibration.scale is None
        with pytest.raises(ValueError):
            calibration(center=(5, 5))
        assert calibration.center == [4, 4]

        with pytest.raises(ValueError):
            calibration.detector(pixel_size=0.1, detector_distance=1, units="nm^-1")

    def test_set_detector(self, calibration):
        calibration.detector(
            pixel_size=15e-6,  # 15 um
            detector_distance=3.8e-2,  # 38 mm
            beam_energy=200,  # 200 keV
            units="k_nm^-1",
        )
        assert not isinstance(calibration.signal.axes_manager[0], UniformDataAxis)
        diff_arr = np.diff(
            calibration.signal.axes_manager[0].axis
        )  # assume mostly flat.
        assert np.allclose(
            diff_arr,
            diff_arr[0],
        )
        assert calibration.flat_ewald == False

    def test_get_slices2d(self, calibration):
        calibration(scale=0.01)
        slices, factors, _, _ = calibration.get_slices2d(5, 90)
        assert len(slices) == 5 * 90

    def test_get_slices_and_factors(self):
        s = Diffraction2D(np.zeros((100, 100)))
        s.calibration(scale=0.1, center=None)
        slices, factors, factor_slices = s.calibration._get_slices_and_factors(
            npt=100, npt_azim=360, radial_range=(0, 4), azimuthal_range=(0, 2 * np.pi)
        )
        # check that the number of pixels for each radial slice is the same
        sum_factors = [np.sum(factors[f[0] : f[1]]) for f in factor_slices]
        sum_factors = np.reshape(sum_factors, (360, 100))
        for row in sum_factors:
            print(np.min(row), np.max(row))
            assert np.allclose(row, row[0], atol=1e-2)
        # Check that the total number of pixels accounted for is equal to the area of the circle
        # Up to rounding due to the fact that we are actually finding the area of an n-gon where
        # n = npt_azim
        all_sum = np.sum(sum_factors)
        assert np.allclose(all_sum, 3.1415 * 40**2, atol=1)
        slices, factors, factor_slices = s.calibration._get_slices_and_factors(
            npt=100, npt_azim=360, radial_range=(0, 15), azimuthal_range=(0, 2 * np.pi)
        )
        # check that the number of pixels for each radial slice is the same
        sum_factors = [np.sum(factors[f[0] : f[1]]) for f in factor_slices]
        sum_factors = np.reshape(sum_factors, (360, 100))
        # Check that the total number of pixels accounted for is equal to the area of the circle
        # Up to rounding due to the fact that we are actually finding the area of an n-gon where
        # n = npt_azim
        all_sum = np.sum(sum_factors)
        assert np.allclose(all_sum, 10000, atol=1)

    def test_get_slices_and_factors1d(self):
        s = Diffraction2D(np.zeros((100, 100)))
        s.calibration(scale=0.1, center=None)
        slices, factors, factor_slices, _ = s.calibration.get_slices1d(
            100, radial_range=(0, 4)
        )
        # check that the number of pixels for each radial slice is the same
        for i in range(len(factor_slices) - 1):
            sl = factors[factor_slices[i] : factor_slices[i + 1]]
            print(np.sum(sl))
        np.testing.assert_almost_equal(np.sum(factors), 3.1415 * 40**2, decimal=0)

    def test_to_string(self, calibration):
        assert (
            str(calibration)
            == "Calibration for <Diffraction2D, title: , dimensions: (|10, 10)>, "
            "Ewald sphere: flat, shape: (10, 10), affine: False, mask: False"
        )
        calibration.detector(
            pixel_size=15e-6, detector_distance=3.8e-2, beam_energy=200, units="k_nm^-1"
        )
        assert (
            str(calibration)
            == "Calibration for <Diffraction2D, title: , dimensions: (|10, 10)>, "
            "Ewald sphere: curved, shape: (10, 10), affine: False, mask: False"
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import pyxem.data.dummy_data as dd


class TestDummyDataModule:
    def test_simple_disk_shift(self):
        s = dd.get_disk_shift_simple_test_signal()
        assert hasattr(s, "plot")
        assert not s._lazy

        s = dd.get_disk_shift_simple_test_signal(lazy=True)
        assert s._lazy

    def test_simple_holz_signal(self):
        s = dd.get_holz_simple_test_signal()
        assert hasattr(s, "plot")
        assert not s._lazy

        s = dd.get_holz_simple_test_signal(lazy=True)
        assert s._lazy

    def test_single_ring_diffraction_signal(self):
        s = dd.get_single_ring_diffraction_signal()
        assert hasattr(s, "plot")

    def test_get_simple_beam_shift_signal(self):
        s = dd.get_simple_beam_shift_signal()
        assert hasattr(s, "plot")

    def test_get_holz_heterostructure_test_signal(self):
        s = dd.get_holz_heterostructure_test_signal(probe_size_x=8, probe_size_y=8)
        assert hasattr(s, "plot")
        assert not s._lazy

        s = dd.get_holz_heterostructure_test_signal(
            lazy=True, probe_size_x=8, probe_size_y=8
        )
        assert s._lazy

    def test_get_stripe_pattern_beam_shift_signal(self):
        s = dd.get_stripe_pattern_beam_shift_signal()
        assert hasattr(s, "plot")

    def test_get_magnetic_square_beam_shift_signal(self):
        s = dd.get_magnetic_square_beam_shift_signal()
        assert hasattr(s, "plot")
        _ = dd.get_magnetic_square_beam_shift_signal(add_ramp=True)
        assert hasattr(s, "plot")

    def test_get_dead_pixel_signal(self):
        s = dd.get_dead_pixel_signal(lazy=True)
        assert hasattr(s, "compute")
        s = dd.get_dead_pixel_signal(lazy=False)
        assert hasattr(s, "plot")
        assert (s.data == 0).any()

    def test_get_hot_pixel_signal(self):
        s = dd.get_hot_pixel_signal(lazy=True)
        assert hasattr(s, "compute")
        s = dd.get_hot_pixel_signal(lazy=False)
        assert hasattr(s, "plot")
        assert (s.data == 50000).any()

    def test_get_cbed_signal(self):
        s = dd.get_cbed_signal()
        assert hasattr(s, "plot")

    def test_get_fem_signal(self):
        s = dd.get_fem_signal()
        assert hasattr(s, "plot")

    def test_get_simple_fem_signal(self):
        s = dd.get_simple_fem_signal()
        assert hasattr(s, "plot")

    def test_get_nanobeam_electron_diffraction_signal(self):
        s = dd.get_nanobeam_electron_diffraction_signal()
        assert hasattr(s, "plot")

    def test_get_generic_fem_signal(self):
        s = dd.get_generic_fem_signal(probe_x=2, probe_y=3, image_x=20, image_y=25)
        assert s.axes_manager.shape == (2, 3, 20, 25)
        assert hasattr(s, "plot")

    def test_get_simple_ellipse_signal_peak_array(self):
        s, peak_array = dd.get_simple_ellipse_signal_peak_array()
        s.add_peak_array_as_markers(peak_array=peak_array)
        assert hasattr(s, "plot")
        assert hasattr(s, "axes_manager")
        assert s.data.shape[:2] == peak_array.shape[:2]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Delete the whole module come 1.0.0

"""Tools for radial integration"""


from pyxem.utils._deprecated import deprecated
from pyxem.signals.reduced_intensity1d import (
    _damp_ri_exponential,
    _damp_ri_extrapolate_to_zero,
    _damp_ri_lorch,
    _damp_ri_low_q_region_erfc,
    _damp_ri_updated_lorch,
)


@deprecated(since="0.18.0", removal="1.0.0")
def subtract_pattern(z, pattern, *args, **kwargs):
    """Used by hs.map in the ReducedIntensityGenerator1D to subtract a background
    pattern.

    Parameters
    ----------
    z : np.array
        A np.array to be transformed
    pattern : np.array
        A numpy array of a single line profile of the same resolution
        (same number of pixels) as the signal to be subtracted from.
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """
    return z - pattern


@deprecated(since="0.18.0", removal="1.0.0")
def mask_from_pattern(z, pattern, *args, **kwargs):
    """Used by hs.map in the ReducedIntensityGenerator1D to mask using a
    background pattern.

    Parameters
    ----------
    z : np.array
        A np.array to be transformed
    pattern : np.array
        A numpy array consisting of 0s and 1s in a single line profile
        of the same resolution (same number of pixels) as the signal.
        1s in the signal are kept. 0s are masked (into zeroes)
    mask_threshold : int or float
        An integer or float threshold. Any pixel in the
        mask_pattern with lower intensity is kept, any with
        higher or equal is set to zero.
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """
    return z * pattern


@deprecated(since="0.18.0", removal="1.0.0")
def damp_ri_exponential(z, b, s_scale, s_size, s_offset, *args, **kwargs):
    """Used by hs.map in the ReducedIntensity1D to damp the reduced
    intensity signal to reduce noise in the high s region by a factor of
    exp(-b*(s^2)), where b is the damping parameter.

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    b : float
        The damping parameter.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    return _damp_ri_exponential(z, b, s_scale, s_size, s_offset, *args, **kwargs)


@deprecated(since="0.18.0", removal="1.0.0")
def damp_ri_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs):
    """Used by hs.map in the ReducedIntensity1D to damp the reduced
    intensity signal to reduce noise in the high s region by a factor of
    sin(s*delta) / (s*delta), where delta = pi / s_max. (from Lorch 1969).

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_max : float
        The maximum s value to be used for transformation to PDF.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    return _damp_ri_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs)


@deprecated(since="0.18.0", removal="1.0.0")
def damp_ri_updated_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs):
    """Used by hs.map in the ReducedIntensity1D to damp the reduced
    intensity signal to reduce noise in the high s region by a factor of
    3 / (s*delta)^3 (sin(s*delta)-s*delta(cos(s*delta))),
    where delta = pi / s_max.

    From "Extracting the pair distribution function from white-beam X-ray
    total scattering data", Soper & Barney, (2011).

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_max : float
        The damping parameter, which need not be the maximum scattering
        vector s to be used for the PDF transform.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    return _damp_ri_updated_lorch(z, s_max, s_scale, s_size, s_offset, *args, **kwargs)


@deprecated(since="0.18.0", removal="1.0.0")
def damp_ri_extrapolate_to_zero(z, s_min, s_scale, s_size, s_offset, *args, **kwargs):
    """Used by hs.map in the ReducedIntensity1D to extrapolate the reduced
    intensity signal to zero below s_min.

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    s_min : float
        Value of s below which data is extrapolated to zero.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    return _damp_ri_extrapolate_to_zero(
        z, s_min, s_scale, s_size, s_offset, *args, **kwargs
    )


@deprecated(since="0.18.0", removal="1.0.0")
def damp_ri_low_q_region_erfc(
    z, scale, offset, s_scale, s_size, s_offset, *args, **kwargs
):
    """Used by hs.map in the ReducedIntensity1D to damp the reduced
    intensity signal in the low q region as a correction to central beam
    effects. The reduced intensity profile is damped by
    (erf(scale * s - offset) + 1) / 2

    Parameters
    ----------
    z : np.array
        A reduced intensity np.array to be transformed.
    scale : float
        A scalar multiplier for s in the error function
    offset : float
        A scalar offset affecting the error function.
    scale : float
        The scattering vector calibation of the reduced intensity array.
    size : int
        The size of the reduced intensity signal. (in pixels)
    *args:
        Arguments to be passed to map().
    **kwargs:
        Keyword arguments to be passed to map().
    """

    return _damp_ri_low_q_region_erfc(
        z, scale, offset, s_scale, s_size, s_offset, *args, **kwargs
    )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Example datasets for use when testing functionality.

Some datasets are packaged with the source code while others must be downloaded from the web.
For more test datasets, see Open datasets.

Datasets are placed in a local cache, in the location returned from pooch.os_cache("pyxem") by default.
the location can be overwritten with a global PYXEM_DATA_DIR environment variable.

With every new version of pyxem, a new directory of datasets with the version name
is added to the cache directory. Any old directories are not deleted automatically, and should
then be deleted manually if desired.
"""

from pyxem.data.simulated_tilt import tilt_boundary_data
from pyxem.data.simulated_si import (
    si_phase,
    si_tilt,
    si_grains,
    si_grains_simple,
    si_rotations_line,
)

from pyxem.data.simulated_overlap import simulated_overlap

from pyxem.data.simulated_fe import fe_bcc_phase, fe_fcc_phase, fe_multi_phase_grains
from pyxem.data._data import (
    au_grating,
    pdnip_glass,
    zrnb_precipitate,
    twinned_nanowire,
    sample_with_g,
    mgo_nanocrystals,
    organic_semiconductor,
    cuag_orientations,
    feal_stripes,
    sped_ag,
    pdcusi_insitu,
    au_grating_20cm,
    small_ptychography,
    au_phase,
    zrcual_1,
    zrcual_2,
    zrcual_3,
)
from pyxem.data.simulated_dpc import (
    simulated_stripes,
    simulated_constant_shift_magnitude,
    simulated_pn_junction,
)

__doctest_skip__ = [
    "au_grating",
    "au_grating_20cm",
    "cuag_orientations",
    "pdnip_glass",
    "zrnb_precipitate",
    "twinned_nanowire",
    "sample_with_g",
    "organic_semiconductor",
    "pdcusi_insitu",
    "feal_stripes",
    "sped_ag",
    "mgo_nanocrystals",
    "zrcual_1",
    "zrcual_2",
    "zrcual_3",
    "small_ptychography",
]

from pyxem.data.simulated_strain import simulated_strain

__all__ = [
    "au_grating",
    "pdnip_glass",
    "zrnb_precipitate",
    "twinned_nanowire",
    "sample_with_g",
    "mgo_nanocrystals",
    "tilt_boundary_data",
    "si_phase",
    "si_tilt",
    "si_grains",
    "si_grains_simple",
    "si_rotations_line",
    "simulated_stripes",
    "fe_multi_phase_grains",
    "fe_fcc_phase",
    "fe_bcc_phase",
    "cuag_orientations",
    "organic_semiconductor",
    "feal_stripes",
    "sped_ag",
    "pdcusi_insitu",
    "simulated_constant_shift_magnitude",
    "au_grating_20cm",
    "small_ptychography",
    "au_phase",
    "simulated_overlap",
    "simulated_strain",
    "simulated_pn_junction",
]



"""Utils for Background Subtraction."""

import numpy as np

from scipy.ndimage import gaussian_filter, median_filter
from skimage.filters.rank import mean as rank_mean
from skimage.morphology import square
from pyxem.utils.diffraction import regional_filter


def _subtract_radial_median(frame, center_x=128, center_y=128):
    """Background removal by subtracting median of pixel at the same
    radius from the center.

    Parameters
    ----------
    frame : NumPy 2D array
    center_x : int
    center_y : int

    Returns
    -------
    background_removed : Numpy 2D array

    Examples
    --------
    >>> import pyxem.utils._dask as dt
    >>> s = pxm.data.dummy_data.get_cbed_signal()
    >>> s_rem = dt._background_removal_single_frame_radial_median(s.data[0, 0])
    """

    y, x = np.indices((frame.shape))
    r = np.hypot(x - center_x, y - center_y)
    r = r.astype(int)
    r_flat = r.ravel()
    diff_image_flat = frame.ravel()
    r_median = np.zeros(np.max(r) + 1, dtype=np.float64)

    for i in range(len(r_median)):
        if diff_image_flat[r_flat == i].size != 0:
            r_median[i] = np.median(diff_image_flat[r_flat == i])
    image = frame - r_median[r]

    return image


def _subtract_dog(frame, min_sigma=1, max_sigma=55):
    """Background removal using difference of Gaussians.

    Parameters
    ----------
    frame : NumPy 2D array
    min_sigma : float
    max_sigma : float

    Returns
    -------
    background_removed : Numpy 2D array

    Examples
    --------
    >>> import pyxem.utils._dask as dt
    >>> s = pxm.data.dummy_data.dummy_data.get_cbed_signal()
    >>> s_rem = dt._background_removal_single_frame_dog(s.data[0, 0])

    """
    blur_max = gaussian_filter(frame, max_sigma)
    blur_min = gaussian_filter(frame, min_sigma)
    return np.maximum(np.where(blur_min > blur_max, frame, 0) - blur_max, 0)


def _subtract_median(frame, footprint=19):
    """Background removal using median filter.

    Parameters
    ----------
    frame : NumPy 2D array
    footprint : float

    Returns
    -------
    background_removed : Numpy 2D array

    Examples
    --------
    >>> import pyxem.utils._dask as dt
    >>> s = pxm.data.dummy_data.get_cbed_signal()
    >>> s_rem = dt._background_removal_single_frame_median(s.data[0, 0])

    """
    bg_subtracted = frame - median_filter(frame, size=footprint)
    return bg_subtracted


def _subtract_hdome(frame, **kwargs):
    """Background removal using h-dome filter."""
    max_value = np.max(frame)
    bg_subtracted = rank_mean(
        regional_filter(frame / max_value, **kwargs), footprint=square(3)
    )
    bg_subtracted = bg_subtracted / np.max(bg_subtracted)
    return bg_subtracted


def _polar_subtract_radial_median(frame):
    """Background removal using the radial median"""
    median = np.nanmedian(frame, axis=1)
    image = frame - median[:, np.newaxis]
    image[image < 0] = 0
    return image


def _polar_subtract_radial_percentile(frame, percentile: int):
    """Background removal using the specified radial percentile.

    Parameters
    ----------
    frame : NumPy 2D array
    percentile : percentile, between 0 and 100

    Note
    -------
    if `percentile` is 50, then this is equivalent to `_polar_subtract_radial_median`.
    """
    percentile = np.nanpercentile(frame, percentile, axis=1)
    image = frame - percentile[:, np.newaxis]
    image[image < 0] = 0
    return image



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Electron diffraction pattern calibration operations."""

import numpy as np
from scipy.optimize import curve_fit
from math import sin, cos
import matplotlib.pyplot as plt

from hyperspy.roi import CircleROI, Line2DROI
from hyperspy.misc.utils import stack as stack_method
from diffsims.utils.ring_pattern_utils import (
    call_ring_pattern,
    calc_radius_with_distortion,
    generate_ring_pattern,
)

from pyxem.utils._deprecated import deprecated
from pyxem.signals import ElectronDiffraction2D


class CalibrationGenerator:
    """Obtains calibration information from common reference standards.

    Parameters
    ----------
    calibration_data : CalibrationDataLibrary
        The signal of electron diffraction data to be used for calibration.

    """

    @deprecated(
        since="0.18.0",
        removal="1.0.0",
        alternative="pyxem.signals.Diffraction2D.calibration",
    )
    def __init__(
        self, diffraction_pattern=None, grating_image=None, calibration_standard=None
    ):
        """
        Parameters
        ------------
        diffraction_pattern : hyperspy.Signal2D
            Some 2 dimensional signal or numpy array of some
            standard sample used for calibration
        grating_image : array_like
            Some 2 dimensional signal or numpy array of some
            standard sample used for calibration
        calibration_standard : diffpy.structure.Structure
            for calculating the polycrystalline ring spacing
        """
        # Assign attributes
        self.diffraction_pattern = diffraction_pattern
        self.grating_image = grating_image
        self.calibration_standard = calibration_standard
        self.mask = None
        self.ring_params = None
        self.affine_matrix = None
        self.rotation_angle = None
        self.correction_matrix = None
        self.center = None
        self.diffraction_calibration = None
        self.navigation_calibration = None

    def __str__(self):
        information_string = (
            "\n|Calibration Data|\n"
            + "=================\n"
            + "Affine Matrix: "
            + str(self.affine_matrix)
            + "\n"
            + "Rotation Angle:"
            + str(self.rotation_angle)
            + "\n"
            + "Center: "
            + str(self.center)
        )
        return information_string

    def get_elliptical_distortion(
        self,
        mask_radius,
        scale=100,
        amplitude=1000,
        spread=2,
        direct_beam_amplitude=500,
        asymmetry=1,
        rotation=0,
        center=None,
    ):
        """Determine elliptical distortion of the diffraction pattern.

        Parameters
        ----------
        mask_radius : int
             The radius in pixels for a mask over the direct beam disc
             (the direct beam disc within given radius will be excluded
             from the fit)
        scale : float
            An initial guess for the diffraction calibration
            in 1/Angstrom units
        amplitude : float
            An initial guess for the amplitude of the polycrystalline rings
            in arbitrary units
        spread : float
            An initial guess for the spread within each ring (Gaussian width)
        direct_beam_amplitude : float
            An initial guess for the background intensity from the direct
            beam disc in arbitrary units
        asymmetry : float
            An initial guess for any elliptical asymmetry in the
            pattern (for a perfectly circular pattern asymmetry=1)
        rotation : float
            An initial guess for the rotation of the (elliptical) pattern
            in radians.
        center : None or list
            The center of the diffraction pattern.

        Returns
        -------
        fit_params : np.array()
            Array of fitting parameters. [scale, amplitude, spread,
                                          direct_beam_amplitude, asymmetry,
                                          rotation].
        affine_matrix : np.array()
            Array defining the affine transformation that corrects for lens
            distortions in the diffraction pattern.

        See Also
        --------
            pyxem.utils.calibration_utils.call_ring_pattern

        """
        # Check that necessary calibration data is provided
        if self.diffraction_pattern is None:
            raise ValueError(
                "This method requires a calibration diffraction pattern"
                " to be provided. Please set self.diffraction_pattern equal"
                " to some Signal2D."
            )
        standard_dp = self.diffraction_pattern
        image_size = standard_dp.data.shape[0]
        if center is None:
            center = [(image_size - 1) / 2, (image_size - 1) / 2]
        # Set diffraction pattern variable
        x, y = np.mgrid[0:image_size, 0:image_size]
        radius_map = calc_radius_with_distortion(x, y, center[0], center[1], 1, 0)
        mask = radius_map < mask_radius
        ref = standard_dp.data[~mask]
        fullx, fully = [x, y]
        pts = np.array([fully[~mask], fullx[~mask]]).ravel()
        # Set initial parameters for fitting
        x0 = [scale, amplitude, spread, direct_beam_amplitude, asymmetry, rotation]
        # Fit ring pattern to experimental data
        xf, cov = curve_fit(call_ring_pattern(center[0], center[1]), pts, ref, p0=x0)
        # Set ring fitting parameters to attribute
        self.ring_params = xf
        # Calculate affine transform parameters from fit parameters
        scaling = np.array([[1, 0], [0, xf[4] ** -0.5]])
        rotation = np.array([[cos(xf[5]), -sin(xf[5])], [sin(xf[5]), cos(xf[5])]])
        correction = np.linalg.inv(np.dot(rotation.T, np.dot(scaling, rotation)))
        affine = np.array(
            [
                [correction[0, 0], correction[0, 1], 0.00],
                [correction[1, 0], correction[1, 1], 0.00],
                [0.00, 0.00, 1.00],
            ]
        )
        # Set affine matrix to attribute
        self.affine_matrix = affine
        return affine

    def get_distortion_residuals(self, mask_radius, spread):
        """Obtain residuals for experimental data and distortion corrected data
        with respect to a simulated symmetric ring pattern.

        Parameters
        ----------
        mask_radius : int
            Radius, in pixels, for a mask over the direct beam disc.
        spread : float
            Gaussian spread of each ring in the simulated pattern.

        Returns
        -------
        diff_init : ElectronDiffraction2D
            Difference between experimental data and simulated symmetric ring
            pattern.
        diff_end : ElectronDiffraction2D
            Difference between distortion corrected data and simulated symmetric
            ring pattern.
        """

        # Check all required parameters are defined as attributes
        if self.diffraction_pattern is None:
            raise ValueError(
                "This method requires an Au X-grating diffraction "
                "pattern to be provided. Please update the "
                "CalibrationDataLibrary."
            )
        if self.affine_matrix is None:
            raise ValueError(
                "This method requires a distortion matrix to have "
                "been determined. Use get_elliptical_distortion "
                "to determine this matrix."
            )
        # Set name for experimental data pattern
        dpeg = self.diffraction_pattern
        ringP = self.ring_params
        size = dpeg.data.shape[0]
        dpref = generate_ring_pattern(
            image_size=size,
            mask=True,
            mask_radius=mask_radius,
            scale=ringP[0],
            amplitude=ringP[1],
            spread=spread,
            direct_beam_amplitude=ringP[3],
            asymmetry=1,
            rotation=ringP[5],
        )
        # Apply distortion corrections to experimental data
        dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
        dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
        dpegs.apply_affine_transformation(
            self.affine_matrix, preserve_range=True, inplace=True
        )
        # Calculate residuals to be returned
        diff_init = ElectronDiffraction2D(dpeg.data - dpref.data)
        diff_end = ElectronDiffraction2D(dpegs.inav[0, 0].data - dpref.data)
        residuals = stack_method([diff_init, diff_end])

        return ElectronDiffraction2D(residuals)

    def plot_corrected_diffraction_pattern(
        self, reference_circle=True, *args, **kwargs
    ):
        """Plot the distortion corrected diffraction pattern with an optional
        reference circle.

        Parameters
        ----------
        reference_circle : bool
            If True a CircleROI widget is added to the plot for reference.
        *args : arguments
            Arguments to be passed to the plot method.
        **kwargs : keyword arguments
            Keyword arguments to be passed to the plot method.

        """
        # Check all required parameters are defined as attributes
        if self.diffraction_pattern is None:
            raise ValueError(
                "This method requires an Au X-grating diffraction "
                "pattern to be provided. Please update the "
                "CalibrationDataLibrary."
            )
        if self.affine_matrix is None:
            raise ValueError(
                "This method requires a distortion matrix to have "
                "been determined. Use get_elliptical_distortion "
                "to determine this matrix."
            )
        # Set name for experimental data pattern
        dpeg = self.diffraction_pattern
        # Apply distortion corrections to experimental data
        size = dpeg.data.shape[0]
        dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
        dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
        dpegs.apply_affine_transformation(
            self.affine_matrix, preserve_range=True, inplace=True
        )
        dpegm = dpegs.mean((0, 1))
        # Plot distortion corrected data
        dpegm.plot(*args, **kwargs)
        # add reference circle if specified
        if reference_circle is True:
            circ = CircleROI(cx=size / 2, cy=size / 2, r=size / 5, r_inner=0)
            circ.add_widget(dpegm)

    def get_diffraction_calibration(self, mask_length, linewidth):
        """Determine the diffraction pattern pixel size calibration in units of
        reciprocal Angstroms per pixel.

        Parameters
        ----------
        mask_length : float
            Halfwidth of the region excluded from peak finding around the
            diffraction pattern center.
        linewidth : float
            Width of Line2DROI used to obtain line trace from distortion
            corrected diffraction pattern.

        Returns
        -------
        diff_cal : float
            Diffraction calibration in reciprocal Angstroms per pixel.

        """
        # Check that necessary calibration data is provided
        if self.diffraction_pattern is None:
            raise ValueError(
                "This method requires an Au X-grating diffraction "
                "pattern to be provided. Please update the "
                "CalibrationDataLibrary."
            )
        if self.affine_matrix is None:
            raise ValueError(
                "This method requires a distortion matrix to have "
                "been determined. Use get_elliptical_distortion "
                "to determine this matrix."
            )
        dpeg = self.diffraction_pattern
        size = dpeg.data.shape[0]
        dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
        dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
        dpegs.apply_affine_transformation(
            self.affine_matrix, preserve_range=True, inplace=True
        )
        dpegm = dpegs.mean((0, 1))
        # Define line roi along which to take trace for calibration
        line = Line2DROI(x1=5, y1=5, x2=size - 6, y2=size - 6, linewidth=linewidth)
        # Obtain line trace
        trace = line(dpegm)
        trace = trace.as_signal1D(0)
        # Find peaks in line trace either side of direct beam
        db = (np.sqrt(2) * (size / 2)) - (5 * np.sqrt(2))
        pka = trace.isig[db + mask_length :].find_peaks1D_ohaver()[0]["position"]
        pkb = trace.isig[: db - mask_length].find_peaks1D_ohaver()[0]["position"]
        # Define Au 220 interplanar spacing (in Angstroms)
        d220 = 1.442
        # Determine predicted position of Au 220 peak
        au_pre = db - (self.ring_params[0] / d220)
        au_post = db + (self.ring_params[0] / d220)
        # Calculate differences between predicted and measured positions
        prediff = np.abs(pkb - au_pre)
        postdiff = np.abs(pka - au_post)
        # Calculate new calibration value based on most accurate peak positions
        dc = (2 / d220) / (
            pka[postdiff == min(postdiff)] - pkb[prediff == min(prediff)]
        )
        # Store diffraction calibration value as attribute
        self.diffraction_calibration = dc[0]

        return dc[0]

    def get_navigation_calibration(self, line_roi, x1, x2, n, xspace, *args, **kwargs):
        """Determine the navigation space pixel size calibration, nm per pixel.

        Parameters
        ----------
        line_roi : Line2DROI
            Line2DROI object along which a profile will be taken to determine
        x1 : float
            Estimate of first X-grating intersection.
        x2 : float
            Estimate of second X-grating intersection.
        n : int
            Number of X-grating squares crossed.
        xspace : float
            Spacing of X-grating in nanometres.
        *args : arguments
            Arguments to be passed to the find_peaks1D method.
        **kwargs : keyword arguments
            Keyword arguments to be passed to the find_peaks1D method.

        Returns
        -------
        nav_cal : float
            Navigation calibration in nanometres per pixel.

        """
        # Check that necessary calibration data is provided
        if self.grating_image is None:
            raise ValueError(
                "This method requires an Au X-grating image to be "
                "provided. Please update the "
                "CalibrationDataLibrary."
            )
        # Obtain line trace
        trace = line_roi(self.grating_image).as_signal1D(0)
        # Find peaks in line trace
        pk = trace.find_peaks1D_ohaver(*args, **kwargs)[0]["position"]
        # Determine peak positions
        dif1 = np.abs(pk - x1)
        dif2 = np.abs(pk - x2)
        # Calculate navigation calibration
        x = (n * xspace) / (pk[dif2 == min(dif2)] - pk[dif1 == min(dif1)])
        # Store navigation calibration value as attribute
        self.navigation_calibration = x[0]
        return x[0]

    def get_rotation_calibration(self, real_line, reciprocal_line):
        """Determine the rotation between real and reciprocal space coordinates.

        Parameters
        ----------
        real_line : Line2DROI
            Line2DROI object drawn along known direction in real space.
        reciprocal_line : Line2DROI
            Line2DROI object drawn along known direction in reciprocal space.

        Returns
        -------
        rotation_angle : float
            Rotation angle in degrees.
        """
        # Calculate rotation angle and store as attribute
        self.rotation_angle = real_line.angle() - reciprocal_line.angle()
        # Return rotation angle calibration
        return self.rotation_angle

    def get_correction_matrix(self):
        """Determine the transformation matrix required to correct for
        diffraction pattern distortions and/or rotation between real and
        reciprocal space coordinates.

        Returns
        -------
        correction_matrix : np.array()
            Array defining the affine transformation that corrects for lens
            distortions in the diffraction pattern.

        """
        if self.affine_matrix is None and self.rotation_angle is None:
            raise ValueError(
                "This method requires either an affine matrix to "
                "correct distortion or a rotation angle between "
                "real and reciprocal space to have been "
                "determined. Please determine these parameters."
            )
        # Only distortion correction case
        elif self.rotation_angle is None:
            correction_matrix = self.affine_matrix
        # Only rotation correction case
        elif self.affine_matrix is None:
            theta = self.rotation_angle / 180 * np.pi
            correction_matrix = np.array(
                [[cos(theta), -sin(theta), 0], [sin(theta), cos(theta), 0], [0, 0, 1]]
            )
        # Otherwise both corrections required, distortion applied first
        else:
            theta = self.rotation_angle / 180 * np.pi
            rotation_matrix = np.array(
                [[cos(theta), -sin(theta), 0], [sin(theta), cos(theta), 0], [0, 0, 1]]
            )
            correction_matrix = np.matmul(rotation_matrix, self.affine_matrix)
        # Set the correction matrix attribute
        self.correction_matrix = correction_matrix
        # Return the correction matrix
        return correction_matrix

    def plot_calibrated_data(
        self, data_to_plot, line=None, unwrap=False, *args, **kwargs
    ):  # pragma: no cover
        """Plot calibrated data for visual inspection.

        Parameters
        ----------
        data_to_plot : str
            Specify the calibrated data to be plotted. Valid options are:
            {'au_x_grating_dp', 'au_x_grating_im', 'moo3_dp', 'moo3_im',
            'rotation_overlay'}
        line : :obj:`hyperspy.roi.Line2DROI`
            An optional Line2DROI object, as detailed in HyperSpy, to be added
            as a widget to the calibration data plot and the trace plotted
            interactively.
        *args : arguments
            Arguments to be passed to the plot method.
        **kwargs : keyword arguments
            Keyword arguments to be passed to the plot method.
        """
        # Construct object containing user defined data to plot and set the
        # calibration checking that it is defined.
        if data_to_plot == "au_x_grating_dp":
            dpeg = self.diffraction_pattern
            size = dpeg.data.shape[0]
            if self.correction_matrix is None:
                self.get_correction_matrix()
            dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
            dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
            dpegs.apply_affine_transformation(
                self.correction_matrix, preserve_range=True, inplace=True
            )
            data = dpegs.mean((0, 1))
            data.set_diffraction_calibration(self.diffraction_calibration)
            # Plot the calibrated diffraction data
            data.plot(*args, **kwargs)
        elif data_to_plot == "au_x_grating_im":
            data = self.diffraction_pattern
            # Plot the calibrated image data
            data.plot(*args, **kwargs)
        elif data_to_plot == "moo3_dp":
            dpeg = self.calibration_data.moo3_dp
            size = dpeg.data.shape[0]
            if self.correction_matrix is None:
                self.get_correction_matrix()
            dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
            dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
            dpegs.apply_affine_transformation(
                self.correction_matrix, preserve_range=True, inplace=True
            )
            data = dpegs.mean((0, 1))
            data.set_diffraction_calibration(self.diffraction_calibration)
            # Plot the calibrated diffraction data
            data.plot(*args, **kwargs)
        elif data_to_plot == "moo3_im":
            data = self.calibration_data.moo3_im
            # Plot the calibrated image data
            data.plot(*args, **kwargs)
        elif data_to_plot == "rotation_overlay":
            dpeg = self.calibration_data.moo3_dp
            size = dpeg.data.shape[0]
            if self.correction_matrix is None:
                self.get_correction_matrix()
            dpegs = stack_method([dpeg, dpeg, dpeg, dpeg])
            dpegs = ElectronDiffraction2D(dpegs.data.reshape((2, 2, size, size)))
            dpegs.apply_affine_transformation(
                self.correction_matrix, preserve_range=True, inplace=True
            )
            dp = dpegs.mean((0, 1))
            im = self.calibration_data.moo3_im.rebin(dp.data.shape)
            stack1 = np.zeros((dp.data.shape[0], dp.data.shape[1], 3))
            stack1[:, :, 0] = dp.data / (0.05 * dp.data.max())
            stack1[:, :, 2] = im.data / im.data.max()
            plt.figure(1)
            plt.imshow(stack1)
        if line:
            line.add_widget(data, axes=data.axes_manager.signal_axes)
            trace = line.interactive(data, navigation_signal="same")
            trace.plot()
            return trace



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import warnings
from hyperspy.signals import Signal1D, Signal2D
from hyperspy._signals.lazy import LazySignal
from pyxem.utils._deprecated import deprecated


class DPCSignal1D(Signal1D):
    """
    Note: this signal will be removed in pyxem version 1.0.0.
    All the functionality has been moved to the BeamShift class.
    Convert this signal to the BeamShift class using the
    DPCSignal.to_beamshift() function.

    Signal for processing differential phase contrast (DPC) acquired using
    scanning transmission electron microscopy (STEM).

    The signal assumes the data is 2 dimensions, where the
    signal dimension is the probe position, and the navigation
    dimension is the x and y disk shifts.

    The first navigation index (s.inav[0]) is assumed to the be x-shift
    and the second navigation is the y-shift (s.inav[1]).

    """

    _signal_type = "dpc"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "DPCSignal1D is deprecated. The functionality has been moved to the "
            "BeamShift class. Use the `to_beamshift()` function to convert to "
            "the BeamShift class.",
            DeprecationWarning,
        )

    def to_beamshift(self):
        """Get BeamShift signal from the DPCSignal.

        The BeamShift signal is a utility signal focused on correcting the shift of the
        center beam in the Diffraction2D signal.

        In practice, the signal and navigation dimensions are switched.

        """
        s_beam_shift = self.T
        s_beam_shift.set_signal_type("beam_shift")
        return s_beam_shift


class DPCSignal2D(Signal2D):
    """
    Note: this signal will be removed in pyxem version 1.0.0.
    All the functionality has been moved to the BeamShift class.
    Convert this signal to the BeamShift class using the
    DPCSignal.to_beamshift() function.

    Signal for processing differential phase contrast (DPC) acquired using
    scanning transmission electron microscopy (STEM).

    The signal assumes the data is 3 dimensions, where the two
    signal dimensions are the probe positions, and the navigation
    dimension is the x and y disk shifts.

    The first navigation index (s.inav[0]) is assumed to the be x-shift
    and the second navigation is the y-shift (s.inav[1]).

    """

    _signal_type = "dpc"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "DPCSignal2D is deprecated. The functionality has been moved to the "
            "BeamShift class. Use the `to_beamshift()` function to convert to "
            "the BeamShift class.",
            DeprecationWarning,
        )

    def to_beamshift(self):
        """Get BeamShift signal from the DPCSignal.

        The BeamShift signal is a utility signal focused on correcting the shift of the
        center beam in the Diffraction2D signal.

        In practice, the signal and navigation dimensions are switched.

        """
        s_beam_shift = self.T
        s_beam_shift.set_signal_type("beam_shift")
        return s_beam_shift


class LazyDPCSignal1D(LazySignal, DPCSignal1D):
    _lazy = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "DPCSignal2D is deprecated. The functionality has been moved to the "
            "BeamShift class. Use the `to_beamshift()` function to convert to "
            "the BeamShift class.",
            DeprecationWarning,
        )


class LazyDPCSignal2D(LazySignal, DPCSignal2D):
    _lazy = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "DPCSignal2D is deprecated. The functionality has been moved to the "
            "BeamShift class. Use the `to_beamshift()` function to convert to "
            "the BeamShift class.",
            DeprecationWarning,
        )
